#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
import sys

from pysnmp import debug
from pysnmp.entity.engine import SnmpEngine
from pysnmp.entity.rfc3413.context import SnmpContext
from pysnmp.proto import errind, error, rfc1902, rfc1905, rfc3411
from pysnmp.proto.api import v2c  # backend is always SMIv2 compliant
from pysnmp.proto.proxy import rfc2576
from pysnmp.smi import error as smi_error


# 3.2
class CommandResponderBase:
    """SNMP command responder base class."""

    ACM_ID = 3  # default MIB access control method to use
    SUPPORTED_PDU_TYPES = ()

    SMI_ERROR_MAP = {
        smi_error.TooBigError: "tooBig",
        smi_error.NoSuchNameError: "noSuchName",
        smi_error.BadValueError: "badValue",
        smi_error.ReadOnlyError: "readOnly",
        smi_error.GenError: "genErr",
        smi_error.NoAccessError: "noAccess",
        smi_error.WrongTypeError: "wrongType",
        smi_error.WrongLengthError: "wrongLength",
        smi_error.WrongEncodingError: "wrongEncoding",
        smi_error.WrongValueError: "wrongValue",
        smi_error.NoCreationError: "noCreation",
        smi_error.InconsistentValueError: "inconsistentValue",
        smi_error.ResourceUnavailableError: "resourceUnavailable",
        smi_error.CommitFailedError: "commitFailed",
        smi_error.UndoFailedError: "undoFailed",
        smi_error.AuthorizationError: "authorizationError",
        smi_error.NotWritableError: "notWritable",
        smi_error.InconsistentNameError: "inconsistentName",
    }

    def __init__(self, snmpEngine: SnmpEngine, snmpContext: SnmpContext, cbCtx=None):
        """Create a responder object."""
        snmpEngine.message_dispatcher.register_context_engine_id(
            snmpContext.contextEngineId, self.SUPPORTED_PDU_TYPES, self.process_pdu
        )
        self.snmpContext = snmpContext
        self.cbCtx = cbCtx
        self.__pendingReqs = {}

    def handle_management_operation(
        self, snmpEngine, stateReference, contextName, PDU, acCtx
    ):
        """Handle incoming SNMP PDU."""
        pass

    def close(self, snmpEngine: SnmpEngine):
        """Unregister responder object."""
        snmpEngine.message_dispatcher.unregister_context_engine_id(
            self.snmpContext.contextEngineId, self.SUPPORTED_PDU_TYPES
        )
        self.snmpContext = self.__pendingReqs = None

    def send_varbinds(
        self, snmpEngine: SnmpEngine, stateReference, errorStatus, errorIndex, varBinds
    ):
        """Send VarBinds."""
        (
            messageProcessingModel,
            securityModel,
            securityName,
            securityLevel,
            contextEngineId,
            contextName,
            pduVersion,
            PDU,
            origPdu,
            maxSizeResponseScopedPDU,
            statusInformation,
        ) = self.__pendingReqs[stateReference]

        v2c.apiPDU.set_error_status(PDU, errorStatus)
        v2c.apiPDU.set_error_index(PDU, errorIndex)
        v2c.apiPDU.set_varbinds(PDU, varBinds)

        debug.logger & debug.FLAG_APP and debug.logger(
            "sendVarBinds: stateReference {}, errorStatus {}, errorIndex {}, varBinds {}".format(
                stateReference, errorStatus, errorIndex, varBinds
            )
        )

        self.send_pdu(snmpEngine, stateReference, PDU)

    def send_pdu(self, snmpEngine: SnmpEngine, stateReference, PDU):
        """Send PDU."""
        (
            messageProcessingModel,
            securityModel,
            securityName,
            securityLevel,
            contextEngineId,
            contextName,
            pduVersion,
            _,
            origPdu,
            maxSizeResponseScopedPDU,
            statusInformation,
        ) = self.__pendingReqs[stateReference]

        # Agent-side API complies with SMIv2
        if messageProcessingModel == 0:
            PDU = rfc2576.v2_to_v1(PDU, origPdu)

        # 3.2.6
        try:
            snmpEngine.message_dispatcher.return_response_pdu(
                snmpEngine,
                messageProcessingModel,
                securityModel,
                securityName,
                securityLevel,
                contextEngineId,
                contextName,
                pduVersion,
                PDU,
                maxSizeResponseScopedPDU,
                stateReference,
                statusInformation,
            )

        except error.StatusInformation:
            debug.logger & debug.FLAG_APP and debug.logger(
                f"sendPdu: stateReference {stateReference}, statusInformation {sys.exc_info()[1]}"
            )
            (snmpSilentDrops,) = snmpEngine.get_mib_builder().import_symbols(  # type: ignore
                "__SNMPv2-MIB", "snmpSilentDrops"
            )
            snmpSilentDrops.syntax += 1

    _get_request_type = rfc1905.GetRequestPDU.tagSet
    _next_request_type = rfc1905.GetNextRequestPDU.tagSet
    _set_request_type = rfc1905.SetRequestPDU.tagSet
    _counter64_type = rfc1902.Counter64.tagSet

    def release_state_information(self, stateReference):
        """Release state information."""
        if stateReference in self.__pendingReqs:
            del self.__pendingReqs[stateReference]

    def process_pdu(
        self,
        snmpEngine: SnmpEngine,
        messageProcessingModel,
        securityModel,
        securityName,
        securityLevel,
        contextEngineId,
        contextName,
        pduVersion,
        PDU,
        maxSizeResponseScopedPDU,
        stateReference,
    ):
        """Process incoming PDU."""
        # Agent-side API complies with SMIv2
        if messageProcessingModel == 0:
            origPdu = PDU
            PDU = rfc2576.v1_to_v2(PDU)
        else:
            origPdu = None

        # 3.2.1
        if (
            PDU.tagSet not in rfc3411.READ_CLASS_PDUS
            and PDU.tagSet not in rfc3411.WRITE_CLASS_PDUS
        ):
            raise error.ProtocolError("Unexpected PDU class %s" % PDU.tagSet)

        # 3.2.2 --> no-op

        # 3.2.4
        rspPDU = v2c.apiPDU.get_response(PDU)

        statusInformation = {}

        self.__pendingReqs[stateReference] = (
            messageProcessingModel,
            securityModel,
            securityName,
            securityLevel,
            contextEngineId,
            contextName,
            pduVersion,
            rspPDU,
            origPdu,
            maxSizeResponseScopedPDU,
            statusInformation,
        )

        # 3.2.5
        varBinds = v2c.apiPDU.get_varbinds(PDU)

        debug.logger & debug.FLAG_APP and debug.logger(
            f"processPdu: stateReference {stateReference}, varBinds {varBinds}"
        )

        try:
            self.handle_management_operation(
                snmpEngine, stateReference, contextName, PDU
            )

        # SNMPv2 SMI exceptions
        except smi_error.SmiError:
            errorIndication = sys.exc_info()[1]

            debug.logger & debug.FLAG_APP and debug.logger(
                f"processPdu: stateReference {stateReference}, errorIndication {errorIndication}"
            )
            if "oid" in errorIndication:
                # Request REPORT generation
                statusInformation["oid"] = errorIndication["oid"]
                statusInformation["val"] = errorIndication["val"]

            errorStatus = self.SMI_ERROR_MAP.get(errorIndication.__class__, "genErr")

            try:
                errorIndex = errorIndication["idx"] + 1

            except KeyError:
                errorIndex = 1

            if len(varBinds) > errorIndex:
                errorIndex = 1

            # rfc1905: 4.2.1.3
            if errorStatus == "tooBig":
                errorIndex = 0
                varBinds = []

            # Report error
            self.send_varbinds(
                snmpEngine, stateReference, errorStatus, errorIndex, varBinds
            )

        except smi_error.PySnmpError:
            debug.logger & debug.FLAG_APP and debug.logger(
                "processPdu: stateReference %s, error "
                "%s" % (stateReference, sys.exc_info()[1])
            )

        self.release_state_information(stateReference)

    @classmethod
    def verify_access(cls, viewType, varBind, **context) -> "bool | None":
        """Verify access rights for a single OID-value pair."""
        name, val = varBind

        snmpEngine: SnmpEngine = context["snmpEngine"]

        execCtx = snmpEngine.observer.get_execution_context(
            "rfc3412.receiveMessage:request"
        )
        (securityModel, securityName, securityLevel, contextName, pduType) = (
            execCtx["securityModel"],
            execCtx["securityName"],
            execCtx["securityLevel"],
            execCtx["contextName"],
            execCtx["pdu"].getTagSet(),
        )

        try:
            snmpEngine.access_control_model[cls.ACM_ID].is_access_allowed(
                snmpEngine,
                securityModel,
                securityName,
                securityLevel,
                viewType,
                contextName,
                name,
            )

        # Map ACM errors onto SMI ones
        except error.StatusInformation:
            statusInformation = sys.exc_info()[1]
            debug.logger & debug.FLAG_APP and debug.logger(
                f"__verifyAccess: name {name}, statusInformation {statusInformation}"
            )
            errorIndication = statusInformation["errorIndication"]
            # 3.2.5...
            if (
                errorIndication == errind.noSuchView
                or errorIndication == errind.noAccessEntry
                or errorIndication == errind.noGroupName
            ):
                raise smi_error.AuthorizationError(name=name, idx=context.get("idx"))

            elif errorIndication == errind.otherError:
                raise smi_error.GenError(name=name, idx=context.get("idx"))

            elif errorIndication == errind.noSuchContext:
                (snmpUnknownContexts,) = snmpEngine.get_mib_builder().import_symbols(
                    "__SNMP-TARGET-MIB", "snmpUnknownContexts"
                )
                snmpUnknownContexts.syntax += 1
                # Request REPORT generation
                raise smi_error.GenError(
                    name=name,
                    idx=context.get("idx"),
                    oid=snmpUnknownContexts.name,
                    val=snmpUnknownContexts.syntax,
                )

            elif errorIndication == errind.notInView:
                return True

            else:
                raise error.ProtocolError("Unknown ACM error %s" % errorIndication)
        else:
            # rfc2576: 4.1.2.1
            if (
                securityModel == 1
                and val is not None
                and cls._counter64_type == val.getTagSet()
                and cls._next_request_type == pduType
            ):
                # This will cause MibTree to skip this OID-value
                raise smi_error.NoAccessError(name=name, idx=context.get("idx"))


class GetCommandResponder(CommandResponderBase):
    """SNMP GET command responder."""

    SUPPORTED_PDU_TYPES = (rfc1905.GetRequestPDU.tagSet,)

    # rfc1905: 4.2.1
    def handle_management_operation(
        self, snmpEngine: SnmpEngine, stateReference, contextName, PDU
    ):
        """Handle incoming SNMP GetRequest-PDU."""
        # rfc1905: 4.2.1.1
        mgmtFun = self.snmpContext.get_mib_instrum(contextName).read_variables
        varBinds = v2c.apiPDU.get_varbinds(PDU)

        context = dict(
            snmpEngine=snmpEngine, acFun=self.verify_access, cbCtx=self.cbCtx
        )

        rspVarBinds = mgmtFun(*varBinds, **context)

        self.send_varbinds(snmpEngine, stateReference, 0, 0, rspVarBinds)
        self.release_state_information(stateReference)


class NextCommandResponder(CommandResponderBase):
    """SNMP GETNEXT command responder."""

    SUPPORTED_PDU_TYPES = (rfc1905.GetNextRequestPDU.tagSet,)

    # rfc1905: 4.2.2
    def handle_management_operation(
        self, snmpEngine: SnmpEngine, stateReference, contextName, PDU
    ):
        """Handle incoming SNMP GetNextRequest-PDU."""
        # rfc1905: 4.2.2.1
        mgmtFun = self.snmpContext.get_mib_instrum(contextName).read_next_variables

        varBinds = v2c.apiPDU.get_varbinds(PDU)

        context = dict(
            snmpEngine=snmpEngine, acFun=self.verify_access, cbCtx=self.cbCtx
        )

        while True:
            rspVarBinds = mgmtFun(*varBinds, **context)

            try:
                self.send_varbinds(snmpEngine, stateReference, 0, 0, rspVarBinds)

            except error.StatusInformation:
                idx = sys.exc_info()[1]["idx"]
                varBinds[idx] = (rspVarBinds[idx][0], varBinds[idx][1])
            else:
                break

        self.release_state_information(stateReference)


class BulkCommandResponder(CommandResponderBase):
    """SNMP GETBULK command responder."""

    SUPPORTED_PDU_TYPES = (rfc1905.GetBulkRequestPDU.tagSet,)
    max_varbinds = 64

    # rfc1905: 4.2.3
    def handle_management_operation(
        self, snmpEngine: SnmpEngine, stateReference, contextName, PDU
    ):
        """Handle incoming SNMP GetBulkRequest-PDU."""
        nonRepeaters = v2c.apiBulkPDU.get_non_repeaters(PDU)
        if nonRepeaters < 0:
            nonRepeaters = 0

        maxRepetitions = v2c.apiBulkPDU.get_max_repetitions(PDU)
        if maxRepetitions < 0:
            maxRepetitions = 0

        reqVarBinds = v2c.apiPDU.get_varbinds(PDU)

        N = min(int(nonRepeaters), len(reqVarBinds))
        M = int(maxRepetitions)
        R = max(len(reqVarBinds) - N, 0)

        if R:
            M = min(M, self.max_varbinds // R)

        debug.logger & debug.FLAG_APP and debug.logger(
            "handleMgmtOperation: N %d, M %d, R %d" % (N, M, R)
        )

        mgmtFun = self.snmpContext.get_mib_instrum(contextName).read_next_variables

        context = dict(
            snmpEngine=snmpEngine, acFun=self.verify_access, cbCtx=self.cbCtx
        )

        if N:
            # TODO(etingof): manage all PDU var-binds in a single call
            rspVarBinds = mgmtFun(*reqVarBinds[:N], **context)

        else:
            rspVarBinds = []

        varBinds = reqVarBinds[-R:]

        while M and R:
            rspVarBinds.extend(mgmtFun(*varBinds, **context))
            varBinds = rspVarBinds[-R:]
            M -= 1

        if len(rspVarBinds):
            self.send_varbinds(snmpEngine, stateReference, 0, 0, rspVarBinds)
            self.release_state_information(stateReference)
        else:
            raise smi_error.SmiError()


class SetCommandResponder(CommandResponderBase):
    """SNMP SET command responder."""

    SUPPORTED_PDU_TYPES = (rfc1905.SetRequestPDU.tagSet,)

    # rfc1905: 4.2.5
    def handle_management_operation(
        self, snmpEngine: SnmpEngine, stateReference, contextName, PDU
    ):
        """Handle incoming SNMP SetRequest-PDU."""
        mgmtFun = self.snmpContext.get_mib_instrum(contextName).write_variables

        varBinds = v2c.apiPDU.get_varbinds(PDU)

        instrumError = None

        context = dict(
            snmpEngine=snmpEngine, acFun=self.verify_access, cbCtx=self.cbCtx
        )

        # rfc1905: 4.2.5.1-13
        try:
            rspVarBinds = mgmtFun(*varBinds, **context)

        except (
            smi_error.NoSuchObjectError,
            smi_error.NoSuchInstanceError,
        ):
            instrumError = smi_error.NotWritableError()
            instrumError.update(sys.exc_info()[1])

        else:
            self.send_varbinds(snmpEngine, stateReference, 0, 0, rspVarBinds)

        self.release_state_information(stateReference)

        if instrumError:
            raise instrumError



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
import warnings

from pyasn1.type import univ
from pysnmp import debug, error
from pysnmp.entity.engine import SnmpEngine
from pysnmp.smi.instrum import MibInstrumController


class SnmpContext:
    """Create a context object."""

    context_names: dict[bytes, MibInstrumController]

    def __init__(self, snmpEngine: SnmpEngine, contextEngineId=None):
        """Create a context object."""
        (snmpEngineId,) = snmpEngine.get_mib_builder().import_symbols(  # type: ignore
            "__SNMP-FRAMEWORK-MIB", "snmpEngineID"
        )
        if contextEngineId is None:
            # Default to local snmpEngineId
            self.contextEngineId = snmpEngineId.syntax
        else:
            self.contextEngineId = snmpEngineId.syntax.clone(contextEngineId)
        debug.logger & debug.FLAG_INS and debug.logger(
            f'SnmpContext: contextEngineId "{self.contextEngineId!r}"'
        )
        self.context_names = {
            b"": snmpEngine.message_dispatcher.mib_instrum_controller  # Default name
        }

    def register_context_name(
        self, contextName, mibInstrum: "MibInstrumController | None" = None
    ):
        """Register a context name."""
        contextName = univ.OctetString(contextName).asOctets()
        if contextName in self.context_names:
            raise error.PySnmpError("Duplicate contextName %s" % contextName)
        debug.logger & debug.FLAG_INS and debug.logger(
            f"registerContextName: registered contextName {contextName!r}, mibInstrum {mibInstrum!r}"
        )
        if mibInstrum is None:
            self.context_names[contextName] = self.context_names[b""]
        else:
            self.context_names[contextName] = mibInstrum

    def unregister_context_name(self, contextName):
        """Unregister a context name."""
        contextName = univ.OctetString(contextName).asOctets()
        if contextName in self.context_names:
            debug.logger & debug.FLAG_INS and debug.logger(
                "unregisterContextName: unregistered contextName %r" % contextName
            )
            del self.context_names[contextName]

    def get_mib_instrum(self, contextName: bytes = b"") -> MibInstrumController:
        """Get MIB instrumentation for a context name."""
        contextName = univ.OctetString(contextName).asOctets()
        if contextName not in self.context_names:
            debug.logger & debug.FLAG_INS and debug.logger(
                "getMibInstrum: contextName %r not registered" % contextName
            )
            raise error.PySnmpError("Missing contextName %s" % contextName)
        else:
            debug.logger & debug.FLAG_INS and debug.logger(
                f"getMibInstrum: contextName {contextName!r}, mibInstum {self.context_names[contextName]!r}"
            )
            return self.context_names[contextName]

    # compatibility with legacy code
    # Old to new attribute mapping
    deprecated_attributes = {
        "getMibInstrum": "get_mib_instrum",
    }

    def __getattr__(self, attr: str):
        """Handle deprecated attributes."""
        if new_attr := self.deprecated_attributes.get(attr):
            warnings.warn(
                f"{attr} is deprecated. Please use {new_attr} instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return getattr(self, new_attr)
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{attr}'"
        )



"""
Listen for notifications at IPv4 & IPv6 interfaces
++++++++++++++++++++++++++++++++++++++++++++++++++

Receive SNMP TRAP messages with the following options:

* SNMPv1/SNMPv2c
* with SNMP community "public"
* over IPv4/UDP, listening at 127.0.0.1:162
* over IPv6/UDP, listening at [::1]:162
* print received data on stdout

Either of the following Net-SNMP commands will send notifications to this
receiver:

| $ snmptrap -v1 -c public 127.0.0.1 1.3.6.1.4.1.20408.4.1.1.2 127.0.0.1 1 1 123 1.3.6.1.2.1.1.1.0 s test
| $ snmptrap -v2c -c public ::1 123 1.3.6.1.6.3.1.1.5.1 1.3.6.1.2.1.1.5.0 s test

Notification Receiver below uses two different transports for communication
with Notification Originators - UDP over IPv4 and UDP over IPv6.

"""  #
from pysnmp.carrier.asyncio.dispatch import AsyncioDispatcher
from pysnmp.carrier.asyncio.dgram import udp, udp6
from pyasn1.codec.ber import decoder
from pysnmp.proto import api


# noinspection PyUnusedLocal
def __callback(transportDispatcher, transportDomain, transportAddress, wholeMsg):
    while wholeMsg:
        msgVer = int(api.decodeMessageVersion(wholeMsg))
        if msgVer in api.PROTOCOL_MODULES:
            pMod = api.PROTOCOL_MODULES[msgVer]

        else:
            print("Unsupported SNMP version %s" % msgVer)
            return

        reqMsg, wholeMsg = decoder.decode(
            wholeMsg,
            asn1Spec=pMod.Message(),
        )

        print(
            "Notification message from {}:{}: ".format(
                transportDomain, transportAddress
            )
        )

        reqPDU = pMod.apiMessage.get_pdu(reqMsg)
        if reqPDU.isSameTypeWith(pMod.TrapPDU()):
            if msgVer == api.SNMP_VERSION_1:
                print(
                    "Enterprise: %s"
                    % (pMod.apiTrapPDU.get_enterprise(reqPDU).prettyPrint())
                )
                print(
                    "Agent Address: %s"
                    % (pMod.apiTrapPDU.get_agent_address(reqPDU).prettyPrint())
                )
                print(
                    "Generic Trap: %s"
                    % (pMod.apiTrapPDU.get_generic_trap(reqPDU).prettyPrint())
                )
                print(
                    "Specific Trap: %s"
                    % (pMod.apiTrapPDU.get_specific_trap(reqPDU).prettyPrint())
                )
                print(
                    "Uptime: %s" % (pMod.apiTrapPDU.get_timestamp(reqPDU).prettyPrint())
                )
                varBinds = pMod.apiTrapPDU.get_varbinds(reqPDU)

            else:
                varBinds = pMod.apiPDU.get_varbinds(reqPDU)

            print("Var-binds:")

            for oid, val in varBinds:
                print(f"{oid.prettyPrint()} = {val.prettyPrint()}")

    return wholeMsg


transportDispatcher = AsyncioDispatcher()

transportDispatcher.register_recv_callback(__callback)

# UDP/IPv4
transportDispatcher.register_transport(
    udp.DOMAIN_NAME, udp.UdpAsyncioTransport().open_server_mode(("localhost", 162))
)

# UDP/IPv6
transportDispatcher.register_transport(
    udp6.DOMAIN_NAME, udp6.Udp6AsyncioTransport().open_server_mode(("::1", 162))
)

transportDispatcher.job_started(1)

try:
    print("This program needs to run as root/administrator to monitor port 162.")
    print("Started. Press Ctrl-C to stop")
    # Dispatcher will never finish as job#1 never reaches zero
    transportDispatcher.run_dispatcher()

except KeyboardInterrupt:
    print("Shutting down...")

finally:
    transportDispatcher.close_dispatcher()



import subprocess
import sys
import re


def run_command(command):
    """Run a shell command and return the output."""
    result = subprocess.run(command, shell=True, capture_output=True, text=True)
    if result.returncode != 0:
        print(f"Error: {result.stderr}")
        sys.exit(result.returncode)
    return result.stdout.strip()


def delete_tag(tag):
    """Delete a Git tag locally and remotely."""
    print(f"Deleting local tag '{tag}'...")
    run_command(f"git tag -d {tag}")

    print(f"Deleting remote tag '{tag}'...")
    run_command(f"git push origin :refs/tags/{tag}")


def create_and_push_tag(tag, branch):
    """Create a Git tag at the head of a branch and push it to remote."""
    print(f"Creating tag '{tag}' at the head of branch '{branch}'...")
    run_command(f"git tag {tag} {branch}")

    print(f"Pushing tag '{tag}' to remote...")
    run_command(f"git push origin {tag}")


def get_largest_release_version():
    """Get the largest release version from the branches."""
    branches = run_command("git branch -r")
    release_branches = re.findall(r"release-(\d+\.\d+)", branches)
    if not release_branches:
        print("No release branches found.")
        sys.exit(1)
    largest_version = max(release_branches, key=lambda v: list(map(int, v.split("."))))
    return largest_version


def main():
    if len(sys.argv) > 2:
        print("Usage: python move-tag.py [<version>]")
        sys.exit(1)

    if len(sys.argv) == 2:
        version = sys.argv[1]
    else:
        version = get_largest_release_version()

    tag = f"v{version}"
    branch = f"release-{version}"

    delete_tag(tag)
    create_and_push_tag(tag, branch)
    print(
        f"Tag '{tag}' has been moved to the head of branch '{branch}' and pushed to remote."
    )


if __name__ == "__main__":
    main()



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
(MibScalarInstance,) = mibBuilder.import_symbols("SNMPv2-SMI", "MibScalarInstance")

(
    usmStatsUnsupportedSecLevels,
    usmStatsNotInTimeWindows,
    usmStatsUnknownUserNames,
    usmStatsUnknownEngineIDs,
    usmStatsWrongDigests,
    usmStatsDecryptionErrors,
    usmUserSpinLock,
) = mibBuilder.import_symbols(
    "SNMP-USER-BASED-SM-MIB",
    "usmStatsUnsupportedSecLevels",
    "usmStatsNotInTimeWindows",
    "usmStatsUnknownUserNames",
    "usmStatsUnknownEngineIDs",
    "usmStatsWrongDigests",
    "usmStatsDecryptionErrors",
    "usmUserSpinLock",
)

__usmStatsUnsupportedSecLevels = MibScalarInstance(
    usmStatsUnsupportedSecLevels.name,
    (0,),
    usmStatsUnsupportedSecLevels.syntax.clone(0),
)
__usmStatsNotInTimeWindows = MibScalarInstance(
    usmStatsNotInTimeWindows.name, (0,), usmStatsNotInTimeWindows.syntax.clone(0)
)
__usmStatsUnknownUserNames = MibScalarInstance(
    usmStatsUnknownUserNames.name, (0,), usmStatsUnknownUserNames.syntax.clone(0)
)
__usmStatsUnknownEngineIDs = MibScalarInstance(
    usmStatsUnknownEngineIDs.name, (0,), usmStatsUnknownEngineIDs.syntax.clone(0)
)
__usmStatsWrongDigests = MibScalarInstance(
    usmStatsWrongDigests.name, (0,), usmStatsWrongDigests.syntax.clone(0)
)
__usmStatsDecryptionErrors = MibScalarInstance(
    usmStatsDecryptionErrors.name, (0,), usmStatsDecryptionErrors.syntax.clone(0)
)
__usmUserSpinLock = MibScalarInstance(
    usmUserSpinLock.name, (0,), usmUserSpinLock.syntax.clone(0)
)

mibBuilder.export_symbols(
    "__SNMP-USER-BASED-SM-MIB",
    usmStatsUnsupportedSecLevels=__usmStatsUnsupportedSecLevels,
    usmStatsNotInTimeWindows=__usmStatsNotInTimeWindows,
    usmStatsUnknownUserNames=__usmStatsUnknownUserNames,
    usmStatsUnknownEngineIDs=__usmStatsUnknownEngineIDs,
    usmStatsWrongDigests=__usmStatsWrongDigests,
    usmStatsDecryptionErrors=__usmStatsDecryptionErrors,
    usmUserSpinLock=__usmUserSpinLock,
)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
(MibScalarInstance,) = mibBuilder.import_symbols("SNMPv2-SMI", "MibScalarInstance")

(
    snmpTargetSpinLock,
    snmpUnavailableContexts,
    snmpUnknownContexts,
) = mibBuilder.import_symbols(
    "SNMP-TARGET-MIB",
    "snmpTargetSpinLock",
    "snmpUnavailableContexts",
    "snmpUnknownContexts",
)

__snmpTargetSpinLock = MibScalarInstance(
    snmpTargetSpinLock.name, (0,), snmpTargetSpinLock.syntax.clone(0)
)
__snmpUnavailableContexts = MibScalarInstance(
    snmpUnavailableContexts.name, (0,), snmpUnavailableContexts.syntax.clone(0)
)
__snmpUnknownContexts = MibScalarInstance(
    snmpUnknownContexts.name, (0,), snmpUnknownContexts.syntax.clone(0)
)

mibBuilder.export_symbols(
    "__SNMP-TARGET-MIB",
    snmpTargetSpinLock=__snmpTargetSpinLock,
    snmpUnavailableContexts=__snmpUnavailableContexts,
    snmpUnknownContexts=__snmpUnknownContexts,
)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# ASN.1 source file://asn1/SNMPv2-TC
# Produced by pysmi-1.5.8 at Sat Nov  2 15:25:23 2024
# On host MacBook-Pro.local platform Darwin version 24.1.0 by user lextm
# Using Python version 3.12.0 (main, Nov 14 2023, 23:52:11) [Clang 15.0.0 (clang-1500.0.40.1)]
#
# IMPORTANT: customization
import inspect
import string
import sys


from pyasn1.type import univ
from pyasn1.type.base import Asn1Item

from pysnmp import debug
from pysnmp.smi.error import (
    InconsistentValueError,
    MibOperationError,
    RowCreationWanted,
    RowDestructionWanted,
    SmiError,
)

# Import base ASN.1 objects even if this MIB does not use it

(Integer,
 OctetString,
 ObjectIdentifier) = mibBuilder.import_symbols(
    "ASN1",
    "Integer",
    "OctetString",
    "ObjectIdentifier")

(NamedValues,) = mibBuilder.import_symbols(
    "ASN1-ENUMERATION",
    "NamedValues")
(ConstraintsIntersection,
 ConstraintsUnion,
 SingleValueConstraint,
 ValueRangeConstraint,
 ValueSizeConstraint) = mibBuilder.import_symbols(
    "ASN1-REFINEMENT",
    "ConstraintsIntersection",
    "ConstraintsUnion",
    "SingleValueConstraint",
    "ValueRangeConstraint",
    "ValueSizeConstraint")

# Import SMI symbols from the MIBs this MIB depends on

# (ModuleCompliance,
#  NotificationGroup) = mibBuilder.import_symbols(
#     "SNMPv2-CONF",
#     "ModuleCompliance",
#     "NotificationGroup")

(Bits,
 Counter32,
 Counter64,
 Gauge32,
 Integer32,
 IpAddress,
 ModuleIdentity,
 MibIdentifier,
 NotificationType,
 ObjectIdentity,
 MibScalar,
 MibTable,
 MibTableRow,
 MibTableColumn,
#  ObjectSyntax,  # TODO: MIB document issue
 TimeTicks,
 Unsigned32,
 iso) = mibBuilder.import_symbols(
    "SNMPv2-SMI",
    "Bits",
    "Counter32",
    "Counter64",
    "Gauge32",
    "Integer32",
    "IpAddress",
    "ModuleIdentity",
    "MibIdentifier",
    "NotificationType",
    "ObjectIdentity",
    "MibScalar",
    "MibTable",
    "MibTableRow",
    "MibTableColumn",
    # "ObjectSyntax",
    "TimeTicks",
    "Unsigned32",
    "iso")

# (DisplayString,
#  TextualConvention) = mibBuilder.import_symbols(
#     "SNMPv2-TC",
#     "DisplayString",
#     "TextualConvention")

# IMPORTANT: customization
class TextualConvention:
    displayHint = ""
    status = "current"
    description = ""
    reference = ""
    bits = ()
    __integer = Integer()
    __unsigned32 = Unsigned32()
    __timeticks = TimeTicks()
    __octetString = OctetString()

    def getDisplayHint(self):
        return self.displayHint

    def getStatus(self):
        return self.status

    def getDescription(self):
        return self.description

    def getReference(self):
        return self.reference

    def getValue(self):
        return self.clone()

    def setValue(self, value):
        if value is None:
            value = univ.noValue
        return self.clone(value)

    def prettyOut(self, value):  # override asn1 type method
        """Implements DISPLAY-HINT evaluation"""
        if self.displayHint and (
            self.__integer.isSuperTypeOf(self, matchConstraints=False)
            and not self.getNamedValues()
            or self.__unsigned32.isSuperTypeOf(self, matchConstraints=False)
            or self.__timeticks.isSuperTypeOf(self, matchConstraints=False)
        ):
            _ = lambda t, f=0: (t, f)
            displayHintType, decimalPrecision = _(*self.displayHint.split("-"))
            if displayHintType == "x":
                return "0x%x" % value
            elif displayHintType == "d":
                try:
                    return "%.*f" % (
                        int(decimalPrecision),
                        float(value) / pow(10, int(decimalPrecision)),
                    )
                except Exception:
                    raise SmiError("float evaluation error: %s" % sys.exc_info()[1])
            elif displayHintType == "o":
                return "0%o" % value
            elif displayHintType == "b":
                runningValue = value
                outputValue = ["B"]
                while runningValue:
                    outputValue.insert(0, "%d" % (runningValue & 0x01))
                    runningValue >>= 1
                return "".join(outputValue)
            else:
                raise SmiError(
                    f'Unsupported numeric type spec "{displayHintType}" at {self.__class__.__name__}'
                )
        elif self.displayHint and self.__octetString.isSuperTypeOf(
            self, matchConstraints=False
        ):
            outputValue = ""
            runningValue = OctetString(value).asOctets()
            displayHint = self.displayHint
            while runningValue and displayHint:
                # 1
                if displayHint[0] == "*":
                    repeatIndicator = repeatCount = runningValue[0]
                    displayHint = displayHint[1:]
                    runningValue = runningValue[1:]
                else:
                    repeatCount = 1
                    repeatIndicator = None

                # 2
                octetLength = ""
                while displayHint and displayHint[0] in string.digits:
                    octetLength += displayHint[0]
                    displayHint = displayHint[1:]

                # length is manatory, but people ignore that
                if not octetLength:
                    octetLength = len(runningValue)

                try:
                    octetLength = int(octetLength)
                except Exception:
                    raise SmiError("Bad octet length: %s" % octetLength)

                if not displayHint:
                    raise SmiError("Short octet length: %s" % self.displayHint)

                # 3
                displayFormat = displayHint[0]
                displayHint = displayHint[1:]

                # 4
                if (
                    displayHint
                    and displayHint[0] not in string.digits
                    and displayHint[0] != "*"
                ):
                    displaySep = displayHint[0]
                    displayHint = displayHint[1:]
                else:
                    displaySep = ""

                # 5
                if displayHint and displaySep and repeatIndicator is not None:
                    repeatTerminator = displayHint[0]
                    displaySep = ""
                    displayHint = displayHint[1:]
                else:
                    repeatTerminator = None

                while repeatCount:
                    repeatCount -= 1
                    if displayFormat == "a":
                        outputValue += runningValue[:octetLength].decode(
                            "ascii", "ignore"
                        )
                    elif displayFormat == "t":
                        outputValue += runningValue[:octetLength].decode(
                            "utf-8", "ignore"
                        )
                    elif displayFormat in ("x", "d", "o"):
                        number = 0
                        numberString = runningValue[:octetLength]
                        while numberString:
                            number <<= 8
                            try:
                                number |= numberString[0]
                                numberString = numberString[1:]
                            except Exception:
                                raise SmiError(
                                    "Display format eval failure: %s: %s"
                                    % (numberString, sys.exc_info()[1])
                                )
                        if displayFormat == "x":
                            outputValue += "%02x" % number
                        elif displayFormat == "o":
                            outputValue += "%03o" % number
                        else:
                            outputValue += "%d" % number
                    else:
                        raise SmiError(
                            "Unsupported display format char: %s" % displayFormat
                        )
                    if runningValue and repeatTerminator:
                        outputValue += repeatTerminator
                    runningValue = runningValue[octetLength:]
                if runningValue and displaySep:
                    outputValue += displaySep
                if not displayHint:
                    displayHint = self.displayHint

            return outputValue

        for base in inspect.getmro(self.__class__):
            if not issubclass(base, TextualConvention) and issubclass(base, Asn1Item):
                return base.prettyOut(self, value)

        raise SmiError("TEXTUAL-CONVENTION has no underlying SNMP base type")

    def prettyIn(self, value):  # override asn1 type method
        """Implements DISPLAY-HINT parsing into base SNMP value

        Proper parsing seems impossible due to ambiguities.
        Here we are trying to do our best, but be prepared
        for failures on complicated DISPLAY-HINTs.

        Keep in mind that this parser only works with "text"
        input meaning `unicode` (Py2) or `str` (Py3).
        """
        for base in inspect.getmro(self.__class__):
            if not issubclass(base, TextualConvention) and issubclass(base, Asn1Item):
                break
        else:
            raise SmiError("TEXTUAL-CONVENTION has no underlying SNMP base type")

        if self.displayHint and (
            self.__integer.isSuperTypeOf(self, matchConstraints=False)
            and not self.getNamedValues()
            or self.__unsigned32.isSuperTypeOf(self, matchConstraints=False)
            or self.__timeticks.isSuperTypeOf(self, matchConstraints=False)
        ):
            value = str(value)

            _ = lambda t, f=0: (t, f)
            displayHintType, decimalPrecision = _(*self.displayHint.split("-"))
            if displayHintType == "x" and (
                value.startswith("0x") or value.startswith("-0x")
            ):
                try:
                    if value.startswith("-"):
                        return base.prettyIn(self, -int(value[3:], 16))
                    else:
                        return base.prettyIn(self, int(value[2:], 16))
                except Exception:
                    raise SmiError("integer evaluation error: %s" % sys.exc_info()[1])
            elif displayHintType == "d":
                try:
                    return base.prettyIn(
                        self, int(float(value) * 10 ** int(decimalPrecision))
                    )
                except Exception:
                    raise SmiError("float evaluation error: %s" % sys.exc_info()[1])
            elif displayHintType == "o" and (
                value.startswith("0") or value.startswith("-0")
            ):
                try:
                    return base.prettyIn(self, int(value, 8))
                except Exception:
                    raise SmiError("octal evaluation error: %s" % sys.exc_info()[1])
            elif displayHintType == "b" and (
                value.startswith("B") or value.startswith("-B")
            ):
                negative = value.startswith("-")
                if negative:
                    value = value[2:]
                else:
                    value = value[1:]
                value = [x != "0" and 1 or 0 for x in value]
                binValue = 0
                while value:
                    binValue <<= value[0]
                    del value[0]
                return base.prettyIn(self, binValue)
            else:
                raise SmiError(
                    f'Unsupported numeric type spec "{displayHintType}" at {self.__class__.__name__}'
                )

        elif self.displayHint and self.__octetString.isSuperTypeOf(
            self, matchConstraints=False
        ):
            numBase = {"x": 16, "d": 10, "o": 8}
            numDigits = {
                "x": string.hexdigits.encode("iso-8859-1"),
                "o": string.octdigits.encode("iso-8859-1"),
                "d": string.digits.encode("iso-8859-1"),
            }

            # how do we know if object is initialized with display-hint
            # formatted text? based on "text" input maybe?
            # That boils down to `str` object on Py3 or `unicode` on Py2.
            if isinstance(value, str) and not isinstance(value, bytes):
                value = base.prettyIn(self, value)
            else:
                return base.prettyIn(self, value)

            outputValue = b""
            runningValue = value
            displayHint = self.displayHint

            while runningValue and displayHint:
                # 1 this information is totally lost, just fail explicitly
                if displayHint[0] == "*":
                    raise SmiError(
                        'Can\'t parse "*" in DISPLAY-HINT (%s)'
                        % self.__class__.__name__
                    )

                # 2 this becomes ambiguous when it comes to rendered value
                octetLength = ""
                while displayHint and displayHint[0] in string.digits:
                    octetLength += displayHint[0]
                    displayHint = displayHint[1:]

                # length is mandatory but people ignore that
                if not octetLength:
                    octetLength = len(runningValue)

                try:
                    octetLength = int(octetLength)
                except Exception:
                    raise SmiError("Bad octet length: %s" % octetLength)

                if not displayHint:
                    raise SmiError("Short octet length: %s" % self.displayHint)

                # 3
                displayFormat = displayHint[0]
                displayHint = displayHint[1:]

                # 4 this is the lifesaver -- we could use it as an anchor
                if (
                    displayHint
                    and displayHint[0] not in string.digits
                    and displayHint[0] != "*"
                ):
                    displaySep = displayHint[0]
                    displayHint = displayHint[1:]
                else:
                    displaySep = ""

                # 5 is probably impossible to support

                if displayFormat in ("a", "t"):
                    outputValue += runningValue[:octetLength]
                elif displayFormat in numBase:
                    if displaySep:
                        guessedOctetLength = runningValue.find(
                            displaySep.encode("iso-8859-1")
                        )
                        if guessedOctetLength == -1:
                            guessedOctetLength = len(runningValue)
                    else:
                        for idx in range(len(runningValue)):
                            if runningValue[idx] not in numDigits[displayFormat]:
                                guessedOctetLength = idx
                                break
                        else:
                            guessedOctetLength = len(runningValue)

                    try:
                        num = int(
                            runningValue[:guessedOctetLength].decode("iso-8859-1"),
                            numBase[displayFormat],
                        )
                    except Exception:
                        raise SmiError(
                            "Display format eval failure: %s: %s"
                            % (runningValue[:guessedOctetLength], sys.exc_info()[1])
                        )

                    num_as_bytes = []
                    if num:
                        while num:
                            num_as_bytes.append(num & 0xFF)
                            num >>= 8
                    else:
                        num_as_bytes = [0]

                    while len(num_as_bytes) < octetLength:
                        num_as_bytes.append(0)

                    num_as_bytes.reverse()

                    outputValue += bytes(num_as_bytes)

                    if displaySep:
                        guessedOctetLength += 1

                    octetLength = guessedOctetLength
                else:
                    raise SmiError(
                        "Unsupported display format char: %s" % displayFormat
                    )

                runningValue = runningValue[octetLength:]

                if not displayHint:
                    displayHint = self.displayHint

            return base.prettyIn(self, outputValue)

        else:
            return base.prettyIn(self, value)



# MODULE-IDENTITY


# Types definitions


# TEXTUAL-CONVENTIONS



# MODULE-IDENTITY


# Types definitions


# TEXTUAL-CONVENTIONS



class DisplayString(TextualConvention, OctetString):
    status = "current"
    displayHint = "255a"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(0, 255),
    )

    if mibBuilder.loadTexts:
        description = "Represents textual information taken from the NVT ASCII character set, as defined in pages 4, 10-11 of RFC 854. To summarize RFC 854, the NVT ASCII repertoire specifies: - the use of character codes 0-127 (decimal) - the graphics characters (32-126) are interpreted as US ASCII - NUL, LF, CR, BEL, BS, HT, VT and FF have the special meanings specified in RFC 854 - the other 25 codes have no standard interpretation - the sequence 'CR LF' means newline - the sequence 'CR NUL' means carriage-return - an 'LF' not preceded by a 'CR' means moving to the same column on the next line. - the sequence 'CR x' for any x other than LF or NUL is illegal. (Note that this also means that a string may end with either 'CR LF' or 'CR NUL', but not with CR.) Any object defined using this syntax may not exceed 255 characters in length."


class PhysAddress(TextualConvention, OctetString):
    status = "current"
    displayHint = "1x:"
    if mibBuilder.loadTexts:
        description = "Represents media- or physical-level addresses."


class MacAddress(TextualConvention, OctetString):
    status = "current"
    displayHint = "1x:"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(6, 6),
    )
    fixed_length = 6

    if mibBuilder.loadTexts:
        description = "Represents an 802 MAC address represented in the `canonical' order defined by IEEE 802.1a, i.e., as if it were transmitted least significant bit first, even though 802.5 (in contrast to other 802.x protocols) requires MAC addresses to be transmitted most significant bit first."


class TruthValue(TextualConvention, Integer32):
    status = "current"
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(1,
              2)
        )
    )
    namedValues = NamedValues(
        *(("true", 1),
          ("false", 2))
    )

    if mibBuilder.loadTexts:
        description = "Represents a boolean value."


class TestAndIncr(TextualConvention, Integer32):
    status = "current"
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueRangeConstraint(0, 2147483647),
    )

    if mibBuilder.loadTexts:
        description = "Represents integer-valued information used for atomic operations. When the management protocol is used to specify that an object instance having this syntax is to be modified, the new value supplied via the management protocol must precisely match the value presently held by the instance. If not, the management protocol set operation fails with an error of `inconsistentValue'. Otherwise, if the current value is the maximum value of 2^31-1 (2147483647 decimal), then the value held by the instance is wrapped to zero; otherwise, the value held by the instance is incremented by one. (Note that regardless of whether the management protocol set operation succeeds, the variable- binding in the request and response PDUs are identical.) The value of the ACCESS clause for objects having this syntax is either `read-write' or `read-create'. When an instance of a columnar object having this syntax is created, any value may be supplied via the management protocol. When the network management portion of the system is re- initialized, the value of every object instance having this syntax must either be incremented from its value prior to the re-initialization, or (if the value prior to the re- initialization is unknown) be set to a pseudo-randomly generated value."

    defaultValue = 0

    def setValue(self, value):
        if value is not None:
            if value != self:
                raise InconsistentValueError()
            value += 1
            if value > 2147483646:
                value = 0
        if value is None:
            value = univ.noValue
        return self.clone(value)

class AutonomousType(TextualConvention, ObjectIdentifier):
    status = "current"
    if mibBuilder.loadTexts:
        description = "Represents an independently extensible type identification value. It may, for example, indicate a particular sub-tree with further MIB definitions, or define a particular type of protocol or hardware."


class InstancePointer(TextualConvention, ObjectIdentifier):
    status = "obsolete"
    if mibBuilder.loadTexts:
        description = "A pointer to either a specific instance of a MIB object or a conceptual row of a MIB table in the managed device. In the latter case, by convention, it is the name of the particular instance of the first accessible columnar object in the conceptual row. The two uses of this textual convention are replaced by VariablePointer and RowPointer, respectively."


class VariablePointer(TextualConvention, ObjectIdentifier):
    status = "current"
    if mibBuilder.loadTexts:
        description = "A pointer to a specific object instance. For example, sysContact.0 or ifInOctets.3."


class RowPointer(TextualConvention, ObjectIdentifier):
    status = "current"
    if mibBuilder.loadTexts:
        description = "Represents a pointer to a conceptual row. The value is the name of the instance of the first accessible columnar object in the conceptual row. For example, ifIndex.3 would point to the 3rd row in the ifTable (note that if ifIndex were not-accessible, then ifDescr.3 would be used instead)."


class RowStatus(TextualConvention, Integer32):
    status = "current"
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1,
              2,
              3,
              4,
              5,
              6)
        )
    )
    namedValues = NamedValues(
        *(("notExists", 0),
          ("active", 1),
          ("notInService", 2),
          ("notReady", 3),
          ("createAndGo", 4),
          ("createAndWait", 5),
          ("destroy", 6))
    )

    if mibBuilder.loadTexts:
        description = "The RowStatus textual convention is used to manage the creation and deletion of conceptual rows, and is used as the value of the SYNTAX clause for the status column of a conceptual row (as described in Section 7.7.1 of [2].) The status column has six defined values: - `active', which indicates that the conceptual row is available for use by the managed device; - `notInService', which indicates that the conceptual row exists in the agent, but is unavailable for use by the managed device (see NOTE below); - `notReady', which indicates that the conceptual row exists in the agent, but is missing information necessary in order to be available for use by the managed device; - `createAndGo', which is supplied by a management station wishing to create a new instance of a conceptual row and to have its status automatically set to active, making it available for use by the managed device; - `createAndWait', which is supplied by a management station wishing to create a new instance of a conceptual row (but not make it available for use by the managed device); and, - `destroy', which is supplied by a management station wishing to delete all of the instances associated with an existing conceptual row. Whereas five of the six values (all except `notReady') may be specified in a management protocol set operation, only three values will be returned in response to a management protocol retrieval operation: `notReady', `notInService' or `active'. That is, when queried, an existing conceptual row has only three states: it is either available for use by the managed device (the status column has value `active'); it is not available for use by the managed device, though the agent has sufficient information to make it so (the status column has value `notInService'); or, it is not available for use by the managed device, and an attempt to make it so would fail because the agent has insufficient information (the state column has value `notReady'). NOTE WELL This textual convention may be used for a MIB table, irrespective of whether the values of that table's conceptual rows are able to be modified while it is active, or whether its conceptual rows must be taken out of service in order to be modified. That is, it is the responsibility of the DESCRIPTION clause of the status column to specify whether the status column must not be `active' in order for the value of some other column of the same conceptual row to be modified. If such a specification is made, affected columns may be changed by an SNMP set PDU if the RowStatus would not be equal to `active' either immediately before or after processing the PDU. In other words, if the PDU also contained a varbind that would change the RowStatus value, the column in question may be changed if the RowStatus was not equal to `active' as the PDU was received, or if the varbind sets the status to a value other than 'active'. Also note that whenever any elements of a row exist, the RowStatus column must also exist. To summarize the effect of having a conceptual row with a status column having a SYNTAX clause value of RowStatus, consider the following state diagram: STATE +--------------+-----------+-------------+------------- | A | B | C | D | |status col.|status column| |status column | is | is |status column ACTION |does not exist| notReady | notInService| is active --------------+--------------+-----------+-------------+------------- set status |noError ->D|inconsist- |inconsistent-|inconsistent- column to | or | entValue| Value| Value createAndGo |inconsistent- | | | | Value| | | --------------+--------------+-----------+-------------+------------- set status |noError see 1|inconsist- |inconsistent-|inconsistent- column to | or | entValue| Value| Value createAndWait |wrongValue | | | --------------+--------------+-----------+-------------+------------- set status |inconsistent- |inconsist- |noError |noError column to | Value| entValue| | active | | | | | | or | | | | | | | |see 2 ->D| ->D| ->D --------------+--------------+-----------+-------------+------------- set status |inconsistent- |inconsist- |noError |noError ->C column to | Value| entValue| | notInService | | | | | | or | | or | | | | | |see 3 ->C| ->C|wrongValue --------------+--------------+-----------+-------------+------------- set status |noError |noError |noError |noError column to | | | | destroy | ->A| ->A| ->A| ->A --------------+--------------+-----------+-------------+------------- set any other |see 4 |noError |noError |see 5 column to some| | | | value | | see 1| ->C| ->D --------------+--------------+-----------+-------------+------------- (1) goto B or C, depending on information available to the agent. (2) if other variable bindings included in the same PDU, provide values for all columns which are missing but required, then return noError and goto D. (3) if other variable bindings included in the same PDU, provide values for all columns which are missing but required, then return noError and goto C. (4) at the discretion of the agent, the return value may be either: inconsistentName: because the agent does not choose to create such an instance when the corresponding RowStatus instance does not exist, or inconsistentValue: if the supplied value is inconsistent with the state of some other MIB object's value, or noError: because the agent chooses to create the instance. If noError is returned, then the instance of the status column must also be created, and the new state is B or C, depending on the information available to the agent. If inconsistentName or inconsistentValue is returned, the row remains in state A. (5) depending on the MIB definition for the column/table, either noError or inconsistentValue may be returned. NOTE: Other processing of the set request may result in a response other than noError being returned, e.g., wrongValue, noCreation, etc. Conceptual Row Creation There are four potential interactions when creating a conceptual row: selecting an instance-identifier which is not in use; creating the conceptual row; initializing any objects for which the agent does not supply a default; and, making the conceptual row available for use by the managed device. Interaction 1: Selecting an Instance-Identifier The algorithm used to select an instance-identifier varies for each conceptual row. In some cases, the instance- identifier is semantically significant, e.g., the destination address of a route, and a management station selects the instance-identifier according to the semantics. In other cases, the instance-identifier is used solely to distinguish conceptual rows, and a management station without specific knowledge of the conceptual row might examine the instances present in order to determine an unused instance-identifier. (This approach may be used, but it is often highly sub-optimal; however, it is also a questionable practice for a naive management station to attempt conceptual row creation.) Alternately, the MIB module which defines the conceptual row might provide one or more objects which provide assistance in determining an unused instance-identifier. For example, if the conceptual row is indexed by an integer-value, then an object having an integer-valued SYNTAX clause might be defined for such a purpose, allowing a management station to issue a management protocol retrieval operation. In order to avoid unnecessary collisions between competing management stations, `adjacent' retrievals of this object should be different. Finally, the management station could select a pseudo-random number to use as the index. In the event that this index was already in use and an inconsistentValue was returned in response to the management protocol set operation, the management station should simply select a new pseudo-random number and retry the operation. A MIB designer should choose between the two latter algorithms based on the size of the table (and therefore the efficiency of each algorithm). For tables in which a large number of entries are expected, it is recommended that a MIB object be defined that returns an acceptable index for creation. For tables with small numbers of entries, it is recommended that the latter pseudo-random index mechanism be used. Interaction 2: Creating the Conceptual Row Once an unused instance-identifier has been selected, the management station determines if it wishes to create and activate the conceptual row in one transaction or in a negotiated set of interactions. Interaction 2a: Creating and Activating the Conceptual Row The management station must first determine the column requirements, i.e., it must determine those columns for which it must or must not provide values. Depending on the complexity of the table and the management station's knowledge of the agent's capabilities, this determination can be made locally by the management station. Alternately, the management station issues a management protocol get operation to examine all columns in the conceptual row that it wishes to create. In response, for each column, there are three possible outcomes: - a value is returned, indicating that some other management station has already created this conceptual row. We return to interaction 1. - the exception `noSuchInstance' is returned, indicating that the agent implements the object-type associated with this column, and that this column in at least one conceptual row would be accessible in the MIB view used by the retrieval were it to exist. For those columns to which the agent provides read-create access, the `noSuchInstance' exception tells the management station that it should supply a value for this column when the conceptual row is to be created. - the exception `noSuchObject' is returned, indicating that the agent does not implement the object-type associated with this column or that there is no conceptual row for which this column would be accessible in the MIB view used by the retrieval. As such, the management station can not issue any management protocol set operations to create an instance of this column. Once the column requirements have been determined, a management protocol set operation is accordingly issued. This operation also sets the new instance of the status column to `createAndGo'. When the agent processes the set operation, it verifies that it has sufficient information to make the conceptual row available for use by the managed device. The information available to the agent is provided by two sources: the management protocol set operation which creates the conceptual row, and, implementation-specific defaults supplied by the agent (note that an agent must provide implementation-specific defaults for at least those objects which it implements as read-only). If there is sufficient information available, then the conceptual row is created, a `noError' response is returned, the status column is set to `active', and no further interactions are necessary (i.e., interactions 3 and 4 are skipped). If there is insufficient information, then the conceptual row is not created, and the set operation fails with an error of `inconsistentValue'. On this error, the management station can issue a management protocol retrieval operation to determine if this was because it failed to specify a value for a required column, or, because the selected instance of the status column already existed. In the latter case, we return to interaction 1. In the former case, the management station can re-issue the set operation with the additional information, or begin interaction 2 again using `createAndWait' in order to negotiate creation of the conceptual row. NOTE WELL Regardless of the method used to determine the column requirements, it is possible that the management station might deem a column necessary when, in fact, the agent will not allow that particular columnar instance to be created or written. In this case, the management protocol set operation will fail with an error such as `noCreation' or `notWritable'. In this case, the management station decides whether it needs to be able to set a value for that particular columnar instance. If not, the management station re-issues the management protocol set operation, but without setting a value for that particular columnar instance; otherwise, the management station aborts the row creation algorithm. Interaction 2b: Negotiating the Creation of the Conceptual Row The management station issues a management protocol set operation which sets the desired instance of the status column to `createAndWait'. If the agent is unwilling to process a request of this sort, the set operation fails with an error of `wrongValue'. (As a consequence, such an agent must be prepared to accept a single management protocol set operation, i.e., interaction 2a above, containing all of the columns indicated by its column requirements.) Otherwise, the conceptual row is created, a `noError' response is returned, and the status column is immediately set to either `notInService' or `notReady', depending on whether it has sufficient information to make the conceptual row available for use by the managed device. If there is sufficient information available, then the status column is set to `notInService'; otherwise, if there is insufficient information, then the status column is set to `notReady'. Regardless, we proceed to interaction 3. Interaction 3: Initializing non-defaulted Objects The management station must now determine the column requirements. It issues a management protocol get operation to examine all columns in the created conceptual row. In the response, for each column, there are three possible outcomes: - a value is returned, indicating that the agent implements the object-type associated with this column and had sufficient information to provide a value. For those columns to which the agent provides read-create access (and for which the agent allows their values to be changed after their creation), a value return tells the management station that it may issue additional management protocol set operations, if it desires, in order to change the value associated with this column. - the exception `noSuchInstance' is returned, indicating that the agent implements the object-type associated with this column, and that this column in at least one conceptual row would be accessible in the MIB view used by the retrieval were it to exist. However, the agent does not have sufficient information to provide a value, and until a value is provided, the conceptual row may not be made available for use by the managed device. For those columns to which the agent provides read-create access, the `noSuchInstance' exception tells the management station that it must issue additional management protocol set operations, in order to provide a value associated with this column. - the exception `noSuchObject' is returned, indicating that the agent does not implement the object-type associated with this column or that there is no conceptual row for which this column would be accessible in the MIB view used by the retrieval. As such, the management station can not issue any management protocol set operations to create an instance of this column. If the value associated with the status column is `notReady', then the management station must first deal with all `noSuchInstance' columns, if any. Having done so, the value of the status column becomes `notInService', and we proceed to interaction 4. Interaction 4: Making the Conceptual Row Available Once the management station is satisfied with the values associated with the columns of the conceptual row, it issues a management protocol set operation to set the status column to `active'. If the agent has sufficient information to make the conceptual row available for use by the managed device, the management protocol set operation succeeds (a `noError' response is returned). Otherwise, the management protocol set operation fails with an error of `inconsistentValue'. NOTE WELL A conceptual row having a status column with value `notInService' or `notReady' is unavailable to the managed device. As such, it is possible for the managed device to create its own instances during the time between the management protocol set operation which sets the status column to `createAndWait' and the management protocol set operation which sets the status column to `active'. In this case, when the management protocol set operation is issued to set the status column to `active', the values held in the agent supersede those used by the managed device. If the management station is prevented from setting the status column to `active' (e.g., due to management station or network failure) the conceptual row will be left in the `notInService' or `notReady' state, consuming resources indefinitely. The agent must detect conceptual rows that have been in either state for an abnormally long period of time and remove them. It is the responsibility of the DESCRIPTION clause of the status column to indicate what an abnormally long period of time would be. This period of time should be long enough to allow for human response time (including `think time') between the creation of the conceptual row and the setting of the status to `active'. In the absense of such information in the DESCRIPTION clause, it is suggested that this period be approximately 5 minutes in length. This removal action applies not only to newly-created rows, but also to previously active rows which are set to, and left in, the notInService state for a prolonged period exceeding that which is considered normal for such a conceptual row. Conceptual Row Suspension When a conceptual row is `active', the management station may issue a management protocol set operation which sets the instance of the status column to `notInService'. If the agent is unwilling to do so, the set operation fails with an error of `wrongValue'. Otherwise, the conceptual row is taken out of service, and a `noError' response is returned. It is the responsibility of the DESCRIPTION clause of the status column to indicate under what circumstances the status column should be taken out of service (e.g., in order for the value of some other column of the same conceptual row to be modified). Conceptual Row Deletion For deletion of conceptual rows, a management protocol set operation is issued which sets the instance of the status column to `destroy'. This request may be made regardless of the current value of the status column (e.g., it is possible to delete conceptual rows which are either `notReady', `notInService' or `active'.) If the operation succeeds, then all instances associated with the conceptual row are immediately removed."

    # Known row states
    (
        stNotExists,
        stActive,
        stNotInService,
        stNotReady,
        stCreateAndGo,
        stCreateAndWait,
        stDestroy,
    ) = list(range(7))

    # States transition matrix (see RFC-1903)
    stateMatrix = {
        # (new-state, current-state)  ->  (error, new-state)
        (stCreateAndGo, stNotExists): (RowCreationWanted, stActive),
        (stCreateAndGo, stNotReady): (InconsistentValueError, stNotReady),
        (stCreateAndGo, stNotInService): (InconsistentValueError, stNotInService),
        (stCreateAndGo, stActive): (InconsistentValueError, stActive),
        #
        (stCreateAndWait, stNotExists): (RowCreationWanted, stActive),
        (stCreateAndWait, stNotReady): (InconsistentValueError, stNotReady),
        (stCreateAndWait, stNotInService): (InconsistentValueError, stNotInService),
        (stCreateAndWait, stActive): (InconsistentValueError, stActive),
        #
        (stActive, stNotExists): (InconsistentValueError, stNotExists),
        (stActive, stNotReady): (InconsistentValueError, stNotReady),
        (stActive, stNotInService): (None, stActive),
        (stActive, stActive): (None, stActive),
        #
        (stNotInService, stNotExists): (InconsistentValueError, stNotExists),
        (stNotInService, stNotReady): (InconsistentValueError, stNotReady),
        (stNotInService, stNotInService): (None, stNotInService),
        (stNotInService, stActive): (None, stActive),
        #
        (stDestroy, stNotExists): (RowDestructionWanted, stNotExists),
        (stDestroy, stNotReady): (RowDestructionWanted, stNotExists),
        (stDestroy, stNotInService): (RowDestructionWanted, stNotExists),
        (stDestroy, stActive): (RowDestructionWanted, stNotExists),
        # This is used on instantiation
        (stNotExists, stNotExists): (None, stNotExists),
    }

    def setValue(self, value):
        if value is None:
            value = univ.noValue

        value = self.clone(value)

        # Run through states transition matrix,
        # resolve new instance value
        excValue, newState = self.stateMatrix.get(
            (
                value.hasValue() and value or self.stNotExists,
                self.hasValue() and self or self.stNotExists,
            ),
            (MibOperationError, None),
        )

        if newState is None:
            newState = univ.noValue

        newState = self.clone(newState)

        debug.logger & debug.FLAG_INS and debug.logger(
            "RowStatus state change from {!r} to {!r} produced new state {!r}, error indication {!r}".format(
                self, value, newState, excValue
            )
        )

        if excValue is not None:
            excValue = excValue(
                msg="Exception at row state transition from {!r} to {!r} yields state {!r} and exception".format(
                    self, value, newState
                ),
                syntax=newState,
            )
            raise excValue

        return newState


class TimeStamp(TextualConvention, TimeTicks):
    status = "current"
    if mibBuilder.loadTexts:
        description = "The value of the sysUpTime object at which a specific occurrence happened. The specific occurrence must be defined in the description of any object defined using this type."


class TimeInterval(TextualConvention, Integer32):
    status = "current"
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueRangeConstraint(0, 2147483647),
    )

    if mibBuilder.loadTexts:
        description = "A period of time, measured in units of 0.01 seconds."


class DateAndTime(TextualConvention, OctetString):
    status = "current"
    displayHint = "2d-1d-1d,1d:1d:1d.1d,1a1d:1d"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(8, 8),
        ValueSizeConstraint(11, 11),
    )

    if mibBuilder.loadTexts:
        description = "A date-time specification. field octets contents range ----- ------ -------- ----- 1 1-2 year 0..65536 2 3 month 1..12 3 4 day 1..31 4 5 hour 0..23 5 6 minutes 0..59 6 7 seconds 0..60 (use 60 for leap-second) 7 8 deci-seconds 0..9 8 9 direction from UTC '+' / '-' 9 10 hours from UTC 0..11 10 11 minutes from UTC 0..59 For example, Tuesday May 26, 1992 at 1:30:15 PM EDT would be displayed as: 1992-5-26,13:30:15.0,-4:0 Note that if only local time is known, then timezone information (fields 8-10) is not present."


class StorageType(TextualConvention, Integer32):
    status = "current"
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(1,
              2,
              3,
              4,
              5)
        )
    )
    namedValues = NamedValues(
        *(("other", 1),
          ("volatile", 2),
          ("nonVolatile", 3),
          ("permanent", 4),
          ("readOnly", 5))
    )

    if mibBuilder.loadTexts:
        description = "Describes the memory realization of a conceptual row. A row which is volatile(2) is lost upon reboot. A row which is either nonVolatile(3), permanent(4) or readOnly(5), is backed up by stable storage. A row which is permanent(4) can be changed but not deleted. A row which is readOnly(5) cannot be changed nor deleted. If the value of an object with this syntax is either permanent(4) or readOnly(5), it cannot be modified. Conversely, if the value is either other(1), volatile(2) or nonVolatile(3), it cannot be modified to be permanent(4) or readOnly(5). Every usage of this textual convention is required to specify the columnar objects which a permanent(4) row must at a minimum allow to be writable."


class TDomain(TextualConvention, ObjectIdentifier):
    status = "current"
    if mibBuilder.loadTexts:
        description = "Denotes a kind of transport service. Some possible values, such as snmpUDPDomain, are defined in 'Transport Mappings for Version 2 of the Simple Network Management Protocol (SNMPv2)'."


class TAddress(TextualConvention, OctetString):
    status = "current"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(1, 255),
    )

    if mibBuilder.loadTexts:
        description = "Denotes a transport service address. For snmpUDPDomain, a TAddress is 6 octets long, the initial 4 octets containing the IP-address in network-byte order and the last 2 containing the UDP port in network-byte order. Consult 'Transport Mappings for Version 2 of the Simple Network Management Protocol (SNMPv2)' for further information on snmpUDPDomain."


# MIB Managed Objects in the order of their OIDs


# Managed Objects groups


# Notification objects


# Notifications groups


# Agent capabilities


# Module compliance


# Export all MIB objects to the MIB builder

mibBuilder.export_symbols(
    "SNMPv2-TC",
    **{"DisplayString": DisplayString,
       "PhysAddress": PhysAddress,
       "MacAddress": MacAddress,
       "TruthValue": TruthValue,
       "TestAndIncr": TestAndIncr,
       "AutonomousType": AutonomousType,
       "InstancePointer": InstancePointer,
       "VariablePointer": VariablePointer,
       "RowPointer": RowPointer,
       "RowStatus": RowStatus,
       "TimeStamp": TimeStamp,
       "TimeInterval": TimeInterval,
       "DateAndTime": DateAndTime,
       "StorageType": StorageType,
       "TDomain": TDomain,
       "TAddress": TAddress,
       "TextualConvention": TextualConvention}
)



import pytest
from pysnmp.hlapi.v3arch.asyncio import *
from pysnmp.proto.errind import UnknownUserName
from tests.agent_context import AGENT_PORT, AgentContextManager


@pytest.mark.asyncio
async def test_usm_no_auth_no_priv():
    async with AgentContextManager():
        with SnmpEngine() as snmpEngine:
            authData = UsmUserData("usr-none-none")
            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpEngine,
                authData,
                await UdpTransportTarget.create(("localhost", AGENT_PORT), retries=0),
                ContextData(),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysDescr.0"
            assert varBinds[0][1].prettyPrint().startswith("PySNMP engine version")
            isinstance(varBinds[0][1], OctetString)


@pytest.mark.asyncio
async def test_usm_no_auth_no_priv_wrong_user():
    async with AgentContextManager():
        with SnmpEngine() as snmpEngine:
            authData = UsmUserData("usr-none-none-not-exist")
            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpEngine,
                authData,
                await UdpTransportTarget.create(("localhost", AGENT_PORT), retries=0),
                ContextData(),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
            )

            assert isinstance(errorIndication, UnknownUserName)
            assert str(errorIndication) == "Unknown USM user"



"""
Detailed VACM configuration
+++++++++++++++++++++++++++

Serves MIB subtrees under different conditions:

* Respond to SNMPv2c commands
* with SNMP community "public"
* over IPv4/UDP, listening at 127.0.0.1:161
* Serve MIB under non-default contextName `abcd`
* Allow access to `SNMPv2-MIB::system` subtree
* Although deny access to `SNMPv2-MIB::sysUpTime` by a bit mask
* Use partial context name matching (`a`)

This example demonstrates detailed VACM configuration performed via
low-level VACM calls: `addContext`, `addVacmGroup`, `addVacmAccess`
and `addVacmView`. Each function populates one of the tables
defined in `SNMP-VIEW-BASED-ACM-MIB` and used strictly as described
in the above mentioned MIB.

The following Net-SNMP's commands will GET a value at this Agent:

| $ snmpget -v2c -c public 127.0.0.1 SNMPv2-MIB::sysLocation.0

However this command will fail:

| $ snmpget -v2c -c public 127.0.0.1 SNMPv2-MIB::sysUpTime.0

This command will not reveal `SNMPv2-MIB::sysUpTime.0` among other objects:

| $ snmpwalk -v2c -c public 127.0.0.1 SNMPv2-MIB::system
"""  #
from pysnmp.entity import engine, config
from pysnmp.entity.rfc3413 import cmdrsp, context
from pysnmp.carrier.asyncio.dgram import udp

# Create SNMP engine with autogenernated engineID and pre-bound
# to socket transport dispatcher
snmpEngine = engine.SnmpEngine()

# Transport setup

# UDP over IPv4
config.add_transport(
    snmpEngine, udp.DOMAIN_NAME, udp.UdpTransport().open_server_mode(("127.0.0.1", 161))
)

# Register default MIB instrumentation controller with a new SNMP context

contextName = "abcd"

snmpContext = context.SnmpContext(snmpEngine)

snmpContext.register_context_name(
    contextName, snmpEngine.message_dispatcher.mib_instrum_controller
)

# Add new SNMP community name, map it to a new security name and
# SNMP context

securityName = "my-area"
communityName = "public"

config.add_v1_system(
    snmpEngine,
    securityName,
    communityName,
    contextEngineId=snmpContext.contextEngineId,
    contextName=contextName,
)

# VACM configuration settings

securityModel = 2  # SNMPv2c
securityLevel = 1  # noAuthNoPriv

vacmGroup = "my-group"
readViewName = "my-read-view"

# We will match by context name prefix
contextPrefix = contextName[:1]

# Populate SNMP-VIEW-BASED-ACM-MIB::vacmContextTable
config.add_context(snmpEngine, contextName)

# Populate SNMP-VIEW-BASED-ACM-MIB::vacmSecurityToGroupTable
config.add_vacm_group(snmpEngine, vacmGroup, securityModel, securityName)

# Populate SNMP-VIEW-BASED-ACM-MIB::vacmAccessTable
config.add_vacm_access(
    snmpEngine,
    vacmGroup,
    contextPrefix,
    securityModel,
    securityLevel,
    "prefix",
    readViewName,
    "",
    "",
)

# Populate SNMP-VIEW-BASED-ACM-MIB::vacmViewTreeFamilyTable

# Allow the whole system subtree
config.add_vacm_view(
    snmpEngine, readViewName, "included", "1.3.6.1.2.1.1.1", "1.1.1.1.1.1.1.0"
)

# ...but exclude one sub-branch (just one scalar OID)
config.add_vacm_view(
    snmpEngine, readViewName, "excluded", "1.3.6.1.2.1.1.3", "1.1.1.1.1.1.1.1"
)

# Register SNMP Applications at the SNMP engine for particular SNMP context
cmdrsp.GetCommandResponder(snmpEngine, snmpContext)
cmdrsp.SetCommandResponder(snmpEngine, snmpContext)
cmdrsp.NextCommandResponder(snmpEngine, snmpContext)

# Register an imaginary never-ending job to keep I/O dispatcher running forever
snmpEngine.transport_dispatcher.job_started(1)

# Run I/O dispatcher which would receive queries and send responses
try:
    snmpEngine.open_dispatcher()

except Exception:
    snmpEngine.close_dispatcher()
    raise



import pytest
from pysnmp.entity.engine import SnmpEngine
from pysnmp.hlapi.v1arch.asyncio.cmdgen import get_cmd, set_cmd, walk_cmd
from pysnmp.hlapi.v1arch.asyncio.dispatch import SnmpDispatcher
from pysnmp.hlapi.v1arch.asyncio.transport import UdpTransportTarget
from pysnmp.hlapi.v1arch.asyncio.auth import CommunityData
from pysnmp.proto.rfc1902 import Integer, OctetString
from pysnmp.smi import builder, compiler, view
from pysnmp.smi.rfc1902 import ObjectIdentity, ObjectType
from tests.agent_context import AGENT_PORT, AgentContextManager


@pytest.mark.asyncio
async def test_v1_set():
    async with AgentContextManager():
        with SnmpDispatcher() as snmpDispatcher:
            iterator = await set_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysLocation", 0), "Shanghai"),
            )

            errorIndication, errorStatus, errorIndex, varBinds = iterator

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysLocation.0"
            assert varBinds[0][1].prettyPrint() == "Shanghai"
            assert isinstance(varBinds[0][1], OctetString)


@pytest.mark.asyncio
async def test_v1_set_mac_address():
    async with AgentContextManager(enable_custom_objects=True):
        snmpDispatcher = SnmpDispatcher()
        # Step 1: Set up MIB builder and add custom MIB directory
        mibBuilder = builder.MibBuilder()
        compiler.addMibCompiler(mibBuilder)
        mibViewController = view.MibViewController(mibBuilder)

        # Load the custom MIB
        mibBuilder.loadModules("LEXTUDIO-TEST-MIB")
        snmpDispatcher.cache["mibViewController"] = mibViewController

        errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
            snmpDispatcher,
            CommunityData("public", mpModel=0),
            await UdpTransportTarget.create(
                ("localhost", AGENT_PORT), timeout=1, retries=0
            ),
            ObjectType(
                ObjectIdentity("LEXTUDIO-TEST-MIB", "testMacAddress", 0)
            ),  # "1.3.6.1.4.1.60069.9.9.0"
        )
        assert errorIndication is None
        assert errorIndication is None
        assert errorStatus == 0
        assert errorIndex == 0
        assert len(varBinds) == 1
        assert varBinds[0][0].prettyPrint() == "LEXTUDIO-TEST-MIB::testMacAddress.0"
        assert (
            varBinds[0][1].prettyPrint() == "00:11:22:33:44:55"
        )  # IMPORTANT: OCTET STRING Size constraint 6..6

        errorIndication, errorStatus, errorIndex, varBinds = await set_cmd(
            snmpDispatcher,
            CommunityData("public", mpModel=0),
            await UdpTransportTarget.create(("localhost", AGENT_PORT)),
            ObjectType(
                ObjectIdentity("LEXTUDIO-TEST-MIB", "testMacAddress", 0),
                OctetString("00:11:22:33:44:55:66"),  # GitHub issue #141
            ),
        )

        assert errorIndication is None
        assert errorStatus == 3  # bad value
        assert errorIndex == 1
        assert len(varBinds) == 1


@pytest.mark.asyncio
async def test_v1_set_table_creation():
    async with AgentContextManager(enable_table_creation=True):
        with SnmpDispatcher() as snmpDispatcher:
            # Perform a SNMP walk to get all object counts
            objects = walk_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ObjectType(ObjectIdentity("1.3.6")),
            )

            objects_list = [item async for item in objects]
            assert len(objects_list) > 0

            object_counts = len(objects_list)

            errorIndication, errorStatus, errorIndex, varBinds = await set_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ObjectType(
                    ObjectIdentity("1.3.6.6.1.5.2.97.98.99"), OctetString("My value")
                ),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-SMI::dod.6.1.5.2.97.98.99"
            assert varBinds[0][1].prettyPrint() == "My value"
            assert type(varBinds[0][1]).__name__ == "OctetString"

            errorIndication, errorStatus, errorIndex, varBinds = await set_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ObjectType(ObjectIdentity("1.3.6.6.1.5.4.97.98.99"), Integer(4)),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-SMI::dod.6.1.5.4.97.98.99"
            assert varBinds[0][1].prettyPrint() == "1"
            # assert isinstance(varBinds[0][1], Integer)

            # Perform a SNMP walk to get all object counts
            objects = walk_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ObjectType(ObjectIdentity("1.3.6")),
            )

            objects_list = [item async for item in objects]
            assert len(objects_list) > 0

            assert len(objects_list) == object_counts + 4



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# PySNMP MIB module SNMP-NOTIFICATION-MIB (https://www.pysnmp.com/pysnmp)
# ASN.1 source file://asn1/SNMP-NOTIFICATION-MIB
# Produced by pysmi-1.5.8 at Sat Nov  2 15:25:33 2024
# On host MacBook-Pro.local platform Darwin version 24.1.0 by user lextm
# Using Python version 3.12.0 (main, Nov 14 2023, 23:52:11) [Clang 15.0.0 (clang-1500.0.40.1)]
#

# Import base ASN.1 objects even if this MIB does not use it

(Integer,
 OctetString,
 ObjectIdentifier) = mibBuilder.import_symbols(
    "ASN1",
    "Integer",
    "OctetString",
    "ObjectIdentifier")

(NamedValues,) = mibBuilder.import_symbols(
    "ASN1-ENUMERATION",
    "NamedValues")
(ConstraintsIntersection,
 ConstraintsUnion,
 SingleValueConstraint,
 ValueRangeConstraint,
 ValueSizeConstraint) = mibBuilder.import_symbols(
    "ASN1-REFINEMENT",
    "ConstraintsIntersection",
    "ConstraintsUnion",
    "SingleValueConstraint",
    "ValueRangeConstraint",
    "ValueSizeConstraint")

# Import SMI symbols from the MIBs this MIB depends on

(SnmpAdminString,) = mibBuilder.import_symbols(
    "SNMP-FRAMEWORK-MIB",
    "SnmpAdminString")

(SnmpTagValue,
 snmpTargetParamsName) = mibBuilder.import_symbols(
    "SNMP-TARGET-MIB",
    "SnmpTagValue",
    "snmpTargetParamsName")

(ModuleCompliance,
 NotificationGroup,
 ObjectGroup) = mibBuilder.import_symbols(
    "SNMPv2-CONF",
    "ModuleCompliance",
    "NotificationGroup",
    "ObjectGroup")

(Bits,
 Counter32,
 Counter64,
 Gauge32,
 Integer32,
 IpAddress,
 ModuleIdentity,
 MibIdentifier,
 NotificationType,
 ObjectIdentity,
 MibScalar,
 MibTable,
 MibTableRow,
 MibTableColumn,
 TimeTicks,
 Unsigned32,
 iso,
 snmpModules) = mibBuilder.import_symbols(
    "SNMPv2-SMI",
    "Bits",
    "Counter32",
    "Counter64",
    "Gauge32",
    "Integer32",
    "IpAddress",
    "ModuleIdentity",
    "MibIdentifier",
    "NotificationType",
    "ObjectIdentity",
    "MibScalar",
    "MibTable",
    "MibTableRow",
    "MibTableColumn",
    "TimeTicks",
    "Unsigned32",
    "iso",
    "snmpModules")

(DisplayString,
 RowStatus,
 StorageType,
 TextualConvention) = mibBuilder.import_symbols(
    "SNMPv2-TC",
    "DisplayString",
    "RowStatus",
    "StorageType",
    "TextualConvention")


# MODULE-IDENTITY

snmpNotificationMIB = ModuleIdentity(
    (1, 3, 6, 1, 6, 3, 13)
)
if mibBuilder.loadTexts:
    snmpNotificationMIB.setRevisions(
        ("2002-10-14 00:00",
         "1998-08-04 00:00",
         "1997-07-14 00:00")
    )
if mibBuilder.loadTexts:
    snmpNotificationMIB.setLastUpdated("200210140000Z")
if mibBuilder.loadTexts:
    snmpNotificationMIB.setOrganization("IETF SNMPv3 Working Group")
if mibBuilder.loadTexts:
    snmpNotificationMIB.setContactInfo("WG-email: snmpv3@lists.tislabs.com Subscribe: majordomo@lists.tislabs.com In message body: subscribe snmpv3 Co-Chair: Russ Mundy Network Associates Laboratories Postal: 15204 Omega Drive, Suite 300 Rockville, MD 20850-4601 USA EMail: mundy@tislabs.com Phone: +1 301-947-7107 Co-Chair: David Harrington Enterasys Networks Postal: 35 Industrial Way P. O. Box 5004 Rochester, New Hampshire 03866-5005 USA EMail: dbh@enterasys.com Phone: +1 603-337-2614 Co-editor: David B. Levi Nortel Networks Postal: 3505 Kesterwood Drive Knoxville, Tennessee 37918 EMail: dlevi@nortelnetworks.com Phone: +1 865 686 0432 Co-editor: Paul Meyer Secure Computing Corporation Postal: 2675 Long Lake Road Roseville, Minnesota 55113 EMail: paul_meyer@securecomputing.com Phone: +1 651 628 1592 Co-editor: Bob Stewart Retired")
if mibBuilder.loadTexts:
    snmpNotificationMIB.setDescription("This MIB module defines MIB objects which provide mechanisms to remotely configure the parameters used by an SNMP entity for the generation of notifications. Copyright (C) The Internet Society (2002). This version of this MIB module is part of RFC 3413; see the RFC itself for full legal notices. ")


# Types definitions


# TEXTUAL-CONVENTIONS



# MIB Managed Objects in the order of their OIDs

_SnmpNotifyObjects_ObjectIdentity = ObjectIdentity
snmpNotifyObjects = _SnmpNotifyObjects_ObjectIdentity(
    (1, 3, 6, 1, 6, 3, 13, 1)
)
_SnmpNotifyTable_Object = MibTable
snmpNotifyTable = _SnmpNotifyTable_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 1)
)
if mibBuilder.loadTexts:
    snmpNotifyTable.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyTable.setDescription("This table is used to select management targets which should receive notifications, as well as the type of notification which should be sent to each selected management target.")
_SnmpNotifyEntry_Object = MibTableRow
snmpNotifyEntry = _SnmpNotifyEntry_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 1, 1)
)
snmpNotifyEntry.setIndexNames(
    (1, "SNMP-NOTIFICATION-MIB", "snmpNotifyName"),
)
if mibBuilder.loadTexts:
    snmpNotifyEntry.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyEntry.setDescription("An entry in this table selects a set of management targets which should receive notifications, as well as the type of notification which should be sent to each selected management target. Entries in the snmpNotifyTable are created and deleted using the snmpNotifyRowStatus object.")


class _SnmpNotifyName_Type(SnmpAdminString):
    """Custom type snmpNotifyName based on SnmpAdminString"""
    subtypeSpec = SnmpAdminString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(1, 32),
    )


_SnmpNotifyName_Type.__name__ = "SnmpAdminString"
_SnmpNotifyName_Object = MibTableColumn
snmpNotifyName = _SnmpNotifyName_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 1, 1, 1),
    _SnmpNotifyName_Type()
)
snmpNotifyName.setMaxAccess("not-accessible")
if mibBuilder.loadTexts:
    snmpNotifyName.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyName.setDescription("The locally arbitrary, but unique identifier associated with this snmpNotifyEntry.")
_SnmpNotifyTag_Type = SnmpTagValue
_SnmpNotifyTag_Object = MibTableColumn
snmpNotifyTag = _SnmpNotifyTag_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 1, 1, 2),
    _SnmpNotifyTag_Type()
)
snmpNotifyTag.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpNotifyTag.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyTag.setDescription("This object contains a single tag value which is used to select entries in the snmpTargetAddrTable. Any entry in the snmpTargetAddrTable which contains a tag value which is equal to the value of an instance of this object is selected. If this object contains a value of zero length, no entries are selected.")


class _SnmpNotifyType_Type(Integer32):
    """Custom type snmpNotifyType based on Integer32"""
    defaultValue = 1

    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(1,
              2)
        )
    )
    namedValues = NamedValues(
        *(("trap", 1),
          ("inform", 2))
    )


_SnmpNotifyType_Type.__name__ = "Integer32"
_SnmpNotifyType_Object = MibTableColumn
snmpNotifyType = _SnmpNotifyType_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 1, 1, 3),
    _SnmpNotifyType_Type()
)
snmpNotifyType.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpNotifyType.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyType.setDescription("This object determines the type of notification to be generated for entries in the snmpTargetAddrTable selected by the corresponding instance of snmpNotifyTag. This value is only used when generating notifications, and is ignored when using the snmpTargetAddrTable for other purposes. If the value of this object is trap(1), then any messages generated for selected rows will contain Unconfirmed-Class PDUs. If the value of this object is inform(2), then any messages generated for selected rows will contain Confirmed-Class PDUs. Note that if an SNMP entity only supports generation of Unconfirmed-Class PDUs (and not Confirmed-Class PDUs), then this object may be read-only.")


class _SnmpNotifyStorageType_Type(StorageType):
    """Custom type snmpNotifyStorageType based on StorageType"""
    defaultValue = 3


_SnmpNotifyStorageType_Type.__name__ = "StorageType"
_SnmpNotifyStorageType_Object = MibTableColumn
snmpNotifyStorageType = _SnmpNotifyStorageType_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 1, 1, 4),
    _SnmpNotifyStorageType_Type()
)
snmpNotifyStorageType.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpNotifyStorageType.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyStorageType.setDescription("The storage type for this conceptual row. Conceptual rows having the value 'permanent' need not allow write-access to any columnar objects in the row.")
_SnmpNotifyRowStatus_Type = RowStatus
_SnmpNotifyRowStatus_Object = MibTableColumn
snmpNotifyRowStatus = _SnmpNotifyRowStatus_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 1, 1, 5),
    _SnmpNotifyRowStatus_Type()
)
snmpNotifyRowStatus.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpNotifyRowStatus.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyRowStatus.setDescription("The status of this conceptual row. To create a row in this table, a manager must set this object to either createAndGo(4) or createAndWait(5).")
_SnmpNotifyFilterProfileTable_Object = MibTable
snmpNotifyFilterProfileTable = _SnmpNotifyFilterProfileTable_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 2)
)
if mibBuilder.loadTexts:
    snmpNotifyFilterProfileTable.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterProfileTable.setDescription("This table is used to associate a notification filter profile with a particular set of target parameters.")
_SnmpNotifyFilterProfileEntry_Object = MibTableRow
snmpNotifyFilterProfileEntry = _SnmpNotifyFilterProfileEntry_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 2, 1)
)
snmpNotifyFilterProfileEntry.setIndexNames(
    (1, "SNMP-TARGET-MIB", "snmpTargetParamsName"),
)
if mibBuilder.loadTexts:
    snmpNotifyFilterProfileEntry.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterProfileEntry.setDescription("An entry in this table indicates the name of the filter profile to be used when generating notifications using the corresponding entry in the snmpTargetParamsTable. Entries in the snmpNotifyFilterProfileTable are created and deleted using the snmpNotifyFilterProfileRowStatus object.")


class _SnmpNotifyFilterProfileName_Type(SnmpAdminString):
    """Custom type snmpNotifyFilterProfileName based on SnmpAdminString"""
    subtypeSpec = SnmpAdminString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(1, 32),
    )


_SnmpNotifyFilterProfileName_Type.__name__ = "SnmpAdminString"
_SnmpNotifyFilterProfileName_Object = MibTableColumn
snmpNotifyFilterProfileName = _SnmpNotifyFilterProfileName_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 2, 1, 1),
    _SnmpNotifyFilterProfileName_Type()
)
snmpNotifyFilterProfileName.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpNotifyFilterProfileName.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterProfileName.setDescription("The name of the filter profile to be used when generating notifications using the corresponding entry in the snmpTargetAddrTable.")


class _SnmpNotifyFilterProfileStorType_Type(StorageType):
    """Custom type snmpNotifyFilterProfileStorType based on StorageType"""
    defaultValue = 3


_SnmpNotifyFilterProfileStorType_Type.__name__ = "StorageType"
_SnmpNotifyFilterProfileStorType_Object = MibTableColumn
snmpNotifyFilterProfileStorType = _SnmpNotifyFilterProfileStorType_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 2, 1, 2),
    _SnmpNotifyFilterProfileStorType_Type()
)
snmpNotifyFilterProfileStorType.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpNotifyFilterProfileStorType.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterProfileStorType.setDescription("The storage type for this conceptual row. Conceptual rows having the value 'permanent' need not allow write-access to any columnar objects in the row.")
_SnmpNotifyFilterProfileRowStatus_Type = RowStatus
_SnmpNotifyFilterProfileRowStatus_Object = MibTableColumn
snmpNotifyFilterProfileRowStatus = _SnmpNotifyFilterProfileRowStatus_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 2, 1, 3),
    _SnmpNotifyFilterProfileRowStatus_Type()
)
snmpNotifyFilterProfileRowStatus.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpNotifyFilterProfileRowStatus.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterProfileRowStatus.setDescription("The status of this conceptual row. To create a row in this table, a manager must set this object to either createAndGo(4) or createAndWait(5). Until instances of all corresponding columns are appropriately configured, the value of the corresponding instance of the snmpNotifyFilterProfileRowStatus column is 'notReady'. In particular, a newly created row cannot be made active until the corresponding instance of snmpNotifyFilterProfileName has been set.")
_SnmpNotifyFilterTable_Object = MibTable
snmpNotifyFilterTable = _SnmpNotifyFilterTable_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 3)
)
if mibBuilder.loadTexts:
    snmpNotifyFilterTable.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterTable.setDescription("The table of filter profiles. Filter profiles are used to determine whether particular management targets should receive particular notifications. When a notification is generated, it must be compared with the filters associated with each management target which is configured to receive notifications, in order to determine whether it may be sent to each such management target. A more complete discussion of notification filtering can be found in section 6. of [SNMP-APPL].")
_SnmpNotifyFilterEntry_Object = MibTableRow
snmpNotifyFilterEntry = _SnmpNotifyFilterEntry_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 3, 1)
)
snmpNotifyFilterEntry.setIndexNames(
    (0, "SNMP-NOTIFICATION-MIB", "snmpNotifyFilterProfileName"),
    (1, "SNMP-NOTIFICATION-MIB", "snmpNotifyFilterSubtree"),
)
if mibBuilder.loadTexts:
    snmpNotifyFilterEntry.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterEntry.setDescription("An element of a filter profile. Entries in the snmpNotifyFilterTable are created and deleted using the snmpNotifyFilterRowStatus object.")
_SnmpNotifyFilterSubtree_Type = ObjectIdentifier
_SnmpNotifyFilterSubtree_Object = MibTableColumn
snmpNotifyFilterSubtree = _SnmpNotifyFilterSubtree_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 3, 1, 1),
    _SnmpNotifyFilterSubtree_Type()
)
snmpNotifyFilterSubtree.setMaxAccess("not-accessible")
if mibBuilder.loadTexts:
    snmpNotifyFilterSubtree.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterSubtree.setDescription("The MIB subtree which, when combined with the corresponding instance of snmpNotifyFilterMask, defines a family of subtrees which are included in or excluded from the filter profile.")


class _SnmpNotifyFilterMask_Type(OctetString):
    """Custom type snmpNotifyFilterMask based on OctetString"""
    defaultHexValue = ""

    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(0, 16),
    )


_SnmpNotifyFilterMask_Type.__name__ = "OctetString"
_SnmpNotifyFilterMask_Object = MibTableColumn
snmpNotifyFilterMask = _SnmpNotifyFilterMask_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 3, 1, 2),
    _SnmpNotifyFilterMask_Type()
)
snmpNotifyFilterMask.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpNotifyFilterMask.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterMask.setDescription("The bit mask which, in combination with the corresponding instance of snmpNotifyFilterSubtree, defines a family of subtrees which are included in or excluded from the filter profile. Each bit of this bit mask corresponds to a sub-identifier of snmpNotifyFilterSubtree, with the most significant bit of the i-th octet of this octet string value (extended if necessary, see below) corresponding to the (8*i - 7)-th sub-identifier, and the least significant bit of the i-th octet of this octet string corresponding to the (8*i)-th sub-identifier, where i is in the range 1 through 16. Each bit of this bit mask specifies whether or not the corresponding sub-identifiers must match when determining if an OBJECT IDENTIFIER matches this family of filter subtrees; a '1' indicates that an exact match must occur; a '0' indicates 'wild card', i.e., any sub-identifier value matches. Thus, the OBJECT IDENTIFIER X of an object instance is contained in a family of filter subtrees if, for each sub-identifier of the value of snmpNotifyFilterSubtree, either: the i-th bit of snmpNotifyFilterMask is 0, or the i-th sub-identifier of X is equal to the i-th sub-identifier of the value of snmpNotifyFilterSubtree. If the value of this bit mask is M bits long and there are more than M sub-identifiers in the corresponding instance of snmpNotifyFilterSubtree, then the bit mask is extended with 1's to be the required length. Note that when the value of this object is the zero-length string, this extension rule results in a mask of all-1's being used (i.e., no 'wild card'), and the family of filter subtrees is the one subtree uniquely identified by the corresponding instance of snmpNotifyFilterSubtree.")


class _SnmpNotifyFilterType_Type(Integer32):
    """Custom type snmpNotifyFilterType based on Integer32"""
    defaultValue = 1

    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(1,
              2)
        )
    )
    namedValues = NamedValues(
        *(("included", 1),
          ("excluded", 2))
    )


_SnmpNotifyFilterType_Type.__name__ = "Integer32"
_SnmpNotifyFilterType_Object = MibTableColumn
snmpNotifyFilterType = _SnmpNotifyFilterType_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 3, 1, 3),
    _SnmpNotifyFilterType_Type()
)
snmpNotifyFilterType.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpNotifyFilterType.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterType.setDescription("This object indicates whether the family of filter subtrees defined by this entry are included in or excluded from a filter. A more detailed discussion of the use of this object can be found in section 6. of [SNMP-APPL].")


class _SnmpNotifyFilterStorageType_Type(StorageType):
    """Custom type snmpNotifyFilterStorageType based on StorageType"""
    defaultValue = 3


_SnmpNotifyFilterStorageType_Type.__name__ = "StorageType"
_SnmpNotifyFilterStorageType_Object = MibTableColumn
snmpNotifyFilterStorageType = _SnmpNotifyFilterStorageType_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 3, 1, 4),
    _SnmpNotifyFilterStorageType_Type()
)
snmpNotifyFilterStorageType.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpNotifyFilterStorageType.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterStorageType.setDescription("The storage type for this conceptual row. Conceptual rows having the value 'permanent' need not allow write-access to any columnar objects in the row.")
_SnmpNotifyFilterRowStatus_Type = RowStatus
_SnmpNotifyFilterRowStatus_Object = MibTableColumn
snmpNotifyFilterRowStatus = _SnmpNotifyFilterRowStatus_Object(
    (1, 3, 6, 1, 6, 3, 13, 1, 3, 1, 5),
    _SnmpNotifyFilterRowStatus_Type()
)
snmpNotifyFilterRowStatus.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpNotifyFilterRowStatus.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterRowStatus.setDescription("The status of this conceptual row. To create a row in this table, a manager must set this object to either createAndGo(4) or createAndWait(5).")
_SnmpNotifyConformance_ObjectIdentity = ObjectIdentity
snmpNotifyConformance = _SnmpNotifyConformance_ObjectIdentity(
    (1, 3, 6, 1, 6, 3, 13, 3)
)
_SnmpNotifyCompliances_ObjectIdentity = ObjectIdentity
snmpNotifyCompliances = _SnmpNotifyCompliances_ObjectIdentity(
    (1, 3, 6, 1, 6, 3, 13, 3, 1)
)
_SnmpNotifyGroups_ObjectIdentity = ObjectIdentity
snmpNotifyGroups = _SnmpNotifyGroups_ObjectIdentity(
    (1, 3, 6, 1, 6, 3, 13, 3, 2)
)

# Managed Objects groups

snmpNotifyGroup = ObjectGroup(
    (1, 3, 6, 1, 6, 3, 13, 3, 2, 1)
)
snmpNotifyGroup.setObjects(
      *(("SNMP-NOTIFICATION-MIB", "snmpNotifyTag"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyType"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyStorageType"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyRowStatus"))
)
if mibBuilder.loadTexts:
    snmpNotifyGroup.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyGroup.setDescription("A collection of objects for selecting which management targets are used for generating notifications, and the type of notification to be generated for each selected management target.")

snmpNotifyFilterGroup = ObjectGroup(
    (1, 3, 6, 1, 6, 3, 13, 3, 2, 2)
)
snmpNotifyFilterGroup.setObjects(
      *(("SNMP-NOTIFICATION-MIB", "snmpNotifyFilterProfileName"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyFilterProfileStorType"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyFilterProfileRowStatus"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyFilterMask"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyFilterType"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyFilterStorageType"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyFilterRowStatus"))
)
if mibBuilder.loadTexts:
    snmpNotifyFilterGroup.setStatus("current")
if mibBuilder.loadTexts:
    snmpNotifyFilterGroup.setDescription("A collection of objects providing remote configuration of notification filters.")


# Notification objects


# Notifications groups


# Agent capabilities


# Module compliance

snmpNotifyBasicCompliance = ModuleCompliance(
    (1, 3, 6, 1, 6, 3, 13, 3, 1, 1)
)
snmpNotifyBasicCompliance.setObjects(
      *(("SNMP-TARGET-MIB", "snmpTargetBasicGroup"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyGroup"))
)
if mibBuilder.loadTexts:
    snmpNotifyBasicCompliance.setStatus(
        "current"
    )
if mibBuilder.loadTexts:
    snmpNotifyBasicCompliance.setDescription("The compliance statement for minimal SNMP entities which implement only SNMP Unconfirmed-Class notifications and read-create operations on only the snmpTargetAddrTable.")

snmpNotifyBasicFiltersCompliance = ModuleCompliance(
    (1, 3, 6, 1, 6, 3, 13, 3, 1, 2)
)
snmpNotifyBasicFiltersCompliance.setObjects(
      *(("SNMP-TARGET-MIB", "snmpTargetBasicGroup"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyGroup"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyFilterGroup"))
)
if mibBuilder.loadTexts:
    snmpNotifyBasicFiltersCompliance.setStatus(
        "current"
    )
if mibBuilder.loadTexts:
    snmpNotifyBasicFiltersCompliance.setDescription("The compliance statement for SNMP entities which implement SNMP Unconfirmed-Class notifications with filtering, and read-create operations on all related tables.")

snmpNotifyFullCompliance = ModuleCompliance(
    (1, 3, 6, 1, 6, 3, 13, 3, 1, 3)
)
snmpNotifyFullCompliance.setObjects(
      *(("SNMP-TARGET-MIB", "snmpTargetBasicGroup"),
        ("SNMP-TARGET-MIB", "snmpTargetResponseGroup"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyGroup"),
        ("SNMP-NOTIFICATION-MIB", "snmpNotifyFilterGroup"))
)
if mibBuilder.loadTexts:
    snmpNotifyFullCompliance.setStatus(
        "current"
    )
if mibBuilder.loadTexts:
    snmpNotifyFullCompliance.setDescription("The compliance statement for SNMP entities which either implement only SNMP Confirmed-Class notifications, or both SNMP Unconfirmed-Class and Confirmed-Class notifications, plus filtering and read-create operations on all related tables.")


# Export all MIB objects to the MIB builder

mibBuilder.export_symbols(
    "SNMP-NOTIFICATION-MIB",
    **{"snmpNotificationMIB": snmpNotificationMIB,
       "snmpNotifyObjects": snmpNotifyObjects,
       "snmpNotifyTable": snmpNotifyTable,
       "snmpNotifyEntry": snmpNotifyEntry,
       "snmpNotifyName": snmpNotifyName,
       "snmpNotifyTag": snmpNotifyTag,
       "snmpNotifyType": snmpNotifyType,
       "snmpNotifyStorageType": snmpNotifyStorageType,
       "snmpNotifyRowStatus": snmpNotifyRowStatus,
       "snmpNotifyFilterProfileTable": snmpNotifyFilterProfileTable,
       "snmpNotifyFilterProfileEntry": snmpNotifyFilterProfileEntry,
       "snmpNotifyFilterProfileName": snmpNotifyFilterProfileName,
       "snmpNotifyFilterProfileStorType": snmpNotifyFilterProfileStorType,
       "snmpNotifyFilterProfileRowStatus": snmpNotifyFilterProfileRowStatus,
       "snmpNotifyFilterTable": snmpNotifyFilterTable,
       "snmpNotifyFilterEntry": snmpNotifyFilterEntry,
       "snmpNotifyFilterSubtree": snmpNotifyFilterSubtree,
       "snmpNotifyFilterMask": snmpNotifyFilterMask,
       "snmpNotifyFilterType": snmpNotifyFilterType,
       "snmpNotifyFilterStorageType": snmpNotifyFilterStorageType,
       "snmpNotifyFilterRowStatus": snmpNotifyFilterRowStatus,
       "snmpNotifyConformance": snmpNotifyConformance,
       "snmpNotifyCompliances": snmpNotifyCompliances,
       "snmpNotifyBasicCompliance": snmpNotifyBasicCompliance,
       "snmpNotifyBasicFiltersCompliance": snmpNotifyBasicFiltersCompliance,
       "snmpNotifyFullCompliance": snmpNotifyFullCompliance,
       "snmpNotifyGroups": snmpNotifyGroups,
       "snmpNotifyGroup": snmpNotifyGroup,
       "snmpNotifyFilterGroup": snmpNotifyFilterGroup,
       "PYSNMP_MODULE_ID": snmpNotificationMIB}
)



"""
SNMPv3: localized auth and privacy keys
+++++++++++++++++++++++++++++++++++++++

Send SNMP GET request using the following options:

* with SNMPv3, user 'usr-md5-des', MD5 authentication, DES encryption
* use localized auth and privacy keys instead of pass-phrase or master keys
* configure authoritative SNMP engine ID (0x0000000000 can be used as well)
* over IPv4/UDP
* to an Agent at demo.pysnmp.com:161
* for SNMPv2-MIB::sysDescr.0 MIB object instance

Functionally similar to:

| $ snmpget -v3 -l authPriv \
      -u usr-md5-des \
      -e 0x80004fb805636c6f75644dab22cc \
      -3k 0x6b99c475259ef7976cf8d028a3381eeb \
      -3K 0x92b5ef98f0a216885e73944e58c07345 \
      demo.pysnmp.com SNMPv2-MIB::sysDescr.0

"""  #
import asyncio
from pysnmp.hlapi.v3arch.asyncio import *


async def run_snmp_get():
    iterator = await get_cmd(
        SnmpEngine(),
        UsmUserData(
            "usr-md5-des",
            authKey=OctetString(hexValue="6b99c475259ef7976cf8d028a3381eeb"),
            privKey=OctetString(hexValue="92b5ef98f0a216885e73944e58c07345"),
            authKeyType=USM_KEY_TYPE_LOCALIZED,
            privKeyType=USM_KEY_TYPE_LOCALIZED,
            securityEngineId=OctetString(hexValue="80004fb805636c6f75644dab22cc"),
        ),
        await UdpTransportTarget.create(("demo.pysnmp.com", 161)),
        ContextData(),
        ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
    )

    errorIndication, errorStatus, errorIndex, varBinds = iterator

    if errorIndication:
        print(errorIndication)

    elif errorStatus:
        print(
            "{} at {}".format(
                errorStatus.prettyPrint(),
                errorIndex and varBinds[int(errorIndex) - 1][0] or "?",
            )
        )

    else:
        for varBind in varBinds:
            print(" = ".join([x.prettyPrint() for x in varBind]))


asyncio.run(run_snmp_get())



import pytest
from pysnmp.hlapi.v3arch.asyncio import *
from tests.agent_context import AGENT_PORT, AgentContextManager


@pytest.mark.asyncio
async def test_custom_asn1_mib_search_path():
    async with AgentContextManager():
        with SnmpEngine() as snmpEngine:
            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpEngine,
                CommunityData("public"),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ContextData(),
                ObjectType(
                    ObjectIdentity("IF-MIB", "ifInOctets", 1).add_asn1_mib_source(
                        "file:///usr/share/snmp", "https://mibs.pysnmp.com/asn1/@mib@"
                    )
                ),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "IF-MIB::ifInOctets.1"



import pytest
from pysnmp.hlapi.v3arch.asyncio import *
from tests.agent_context import AGENT_PORT, AgentContextManager


@pytest.mark.asyncio
async def test_v1_set():
    async with AgentContextManager():
        with SnmpEngine() as snmpEngine:
            errorIndication, errorStatus, errorIndex, varBinds = await set_cmd(
                snmpEngine,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ContextData(),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysLocation", 0), "Shanghai"),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysLocation.0"
            assert varBinds[0][1].prettyPrint() == "Shanghai"
            assert isinstance(varBinds[0][1], OctetString)


@pytest.mark.asyncio
async def test_v1_set_table_creation():
    async with AgentContextManager(enable_table_creation=True):
        with SnmpEngine() as snmpEngine:
            # Perform a SNMP walk to get all object counts
            objects = walk_cmd(
                snmpEngine,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ContextData(),
                ObjectType(ObjectIdentity("1.3.6")),
            )

            objects_list = [item async for item in objects]
            assert len(objects_list) > 0

            object_counts = len(objects_list)

            errorIndication, errorStatus, errorIndex, varBinds = await set_cmd(
                snmpEngine,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ContextData(),
                ObjectType(
                    ObjectIdentity("1.3.6.6.1.5.2.97.98.99"), OctetString("My value")
                ),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-SMI::dod.6.1.5.2.97.98.99"
            assert varBinds[0][1].prettyPrint() == "My value"
            assert isinstance(varBinds[0][1], OctetString)

            errorIndication, errorStatus, errorIndex, varBinds = await set_cmd(
                snmpEngine,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ContextData(),
                ObjectType(ObjectIdentity("1.3.6.6.1.5.4.97.98.99"), Integer(4)),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-SMI::dod.6.1.5.4.97.98.99"
            assert varBinds[0][1].prettyPrint() == "1"
            # assert isinstance(varBinds[0][1], Integer)

            # Perform a SNMP walk to get all object counts
            objects = walk_cmd(
                snmpEngine,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ContextData(),
                ObjectType(ObjectIdentity("1.3.6")),
            )

            objects_list = [item async for item in objects]
            assert len(objects_list) > 0

            assert len(objects_list) == object_counts + 4



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
import sys

from pyasn1.type import univ
from pysnmp import debug, error, nextid
from pysnmp.entity.engine import SnmpEngine
from pysnmp.entity.rfc3413 import config
from pysnmp.proto import errind
from pysnmp.proto.api import v2c
from pysnmp.proto.error import StatusInformation
from pysnmp.proto.proxy import rfc2576

getNextHandle = nextid.Integer(0x7FFFFFFF)  # noqa: N816


class CommandGenerator:
    """SNMP command generator."""

    _null = univ.Null("")

    def __init__(self, **options):
        """Create a command generator object."""
        self.__options = options
        self.__pendingReqs = {}

    def process_response_pdu(
        self,
        snmpEngine: SnmpEngine,
        messageProcessingModel,
        securityModel,
        securityName,
        securityLevel,
        contextEngineId,
        contextName,
        pduVersion,
        PDU,
        statusInformation,
        sendPduHandle,
        cbCtx,
    ):
        """Process SNMP response."""
        origSendRequestHandle, cbFun, cbCtx = cbCtx

        # 3.1.1
        if sendPduHandle not in self.__pendingReqs:
            raise error.PySnmpError("Missing sendPduHandle %s" % sendPduHandle)

        (
            origTransportDomain,
            origTransportAddress,
            origMessageProcessingModel,
            origSecurityModel,
            origSecurityName,
            origSecurityLevel,
            origContextEngineId,
            origContextName,
            origPduVersion,
            origPdu,
            origTimeout,
            origRetryCount,
            origRetries,
            origDiscoveryRetries,
        ) = self.__pendingReqs.pop(sendPduHandle)

        snmpEngine.transport_dispatcher.job_finished(id(self))

        # 3.1.3
        if statusInformation:
            debug.logger & debug.FLAG_APP and debug.logger(
                f"processResponsePdu: sendPduHandle {sendPduHandle}, statusInformation {statusInformation}"
            )

            errorIndication = statusInformation["errorIndication"]

            if errorIndication in (errind.notInTimeWindow, errind.unknownEngineID):
                origDiscoveryRetries += 1
                origRetries = 0
            else:
                origDiscoveryRetries = 0
                origRetries += 1

            if (
                origRetries > origRetryCount
                or origDiscoveryRetries > self.__options.get("discoveryRetries", 4)
            ):
                debug.logger & debug.FLAG_APP and debug.logger(
                    "processResponsePdu: sendPduHandle %s, retry count %d exceeded"
                    % (sendPduHandle, origRetries)
                )
                cbFun(snmpEngine, origSendRequestHandle, errorIndication, None, cbCtx)
                return

            # User-side API assumes SMIv2
            if origMessageProcessingModel == 0:
                reqPDU = rfc2576.v2_to_v1(origPdu)
                pduVersion = 0
            else:
                reqPDU = origPdu
                pduVersion = 1

            try:
                sendPduHandle = snmpEngine.message_dispatcher.send_pdu(
                    snmpEngine,
                    origTransportDomain,
                    origTransportAddress,
                    origMessageProcessingModel,
                    origSecurityModel,
                    origSecurityName,
                    origSecurityLevel,
                    origContextEngineId,
                    origContextName,
                    pduVersion,
                    reqPDU,
                    True,
                    origTimeout,
                    self.process_response_pdu,
                    (origSendRequestHandle, cbFun, cbCtx),
                )

                snmpEngine.transport_dispatcher.job_started(id(self))

                self.__pendingReqs[sendPduHandle] = (
                    origTransportDomain,
                    origTransportAddress,
                    origMessageProcessingModel,
                    origSecurityModel,
                    origSecurityName,
                    origSecurityLevel,
                    origContextEngineId,
                    origContextName,
                    origPduVersion,
                    origPdu,
                    origTimeout,
                    origRetryCount,
                    origRetries,
                    origDiscoveryRetries,
                )
                return

            except StatusInformation:
                statusInformation = sys.exc_info()[1]
                debug.logger & debug.FLAG_APP and debug.logger(
                    "processResponsePdu: origSendRequestHandle {}, _sendPdu() failed with {!r}".format(
                        sendPduHandle, statusInformation
                    )
                )
                cbFun(
                    snmpEngine,
                    origSendRequestHandle,
                    statusInformation["errorIndication"],  # type: ignore
                    None,
                    cbCtx,
                )
                return

        if (
            origMessageProcessingModel != messageProcessingModel
            or origSecurityModel != securityModel
            or origSecurityName != origSecurityName
            or origContextEngineId
            and origContextEngineId != contextEngineId
            or origContextName
            and origContextName != contextName
            or origPduVersion != pduVersion
        ):
            debug.logger & debug.FLAG_APP and debug.logger(
                "processResponsePdu: sendPduHandle %s, request/response data mismatch"
                % sendPduHandle
            )

            cbFun(snmpEngine, origSendRequestHandle, "badResponse", None, cbCtx)
            return

        # User-side API assumes SMIv2
        if messageProcessingModel == 0:
            PDU = rfc2576.v1_to_v2(PDU, origPdu)

        # 3.1.2
        if v2c.apiPDU.get_request_id(PDU) != v2c.apiPDU.get_request_id(origPdu):
            debug.logger & debug.FLAG_APP and debug.logger(
                "processResponsePdu: sendPduHandle %s, request-id/response-id mismatch"
                % sendPduHandle
            )
            cbFun(snmpEngine, origSendRequestHandle, "badResponse", None, cbCtx)
            return

        cbFun(snmpEngine, origSendRequestHandle, None, PDU, cbCtx)

    def send_pdu(
        self,
        snmpEngine: SnmpEngine,
        targetName,
        contextEngineId,
        contextName,
        PDU,
        cbFun,
        cbCtx,
    ):
        """Send SNMP PDU."""
        (
            transportDomain,
            transportAddress,
            timeout,
            retryCount,
            messageProcessingModel,
            securityModel,
            securityName,
            securityLevel,
        ) = config.get_target_info(snmpEngine, targetName)

        if snmpEngine.transport_dispatcher is None:
            raise error.PySnmpError("No transport dispatcher available")
        # Convert timeout in seconds into timeout in timer ticks
        timeoutInTicks = (
            float(timeout)
            / 100
            / snmpEngine.transport_dispatcher.get_timer_resolution()
        )

        SnmpEngineID, SnmpAdminString = snmpEngine.get_mib_builder().import_symbols(  # type: ignore
            "SNMP-FRAMEWORK-MIB", "SnmpEngineID", "SnmpAdminString"
        )

        # Cast possible strings into bytes
        if contextEngineId:
            contextEngineId = SnmpEngineID(contextEngineId)
        contextName = SnmpAdminString(contextName)

        origPDU = PDU

        # User-side API assumes SMIv2
        if messageProcessingModel == 0:
            PDU = rfc2576.v2_to_v1(PDU)
            pduVersion = 0
        else:
            pduVersion = 1

        sendRequestHandle = getNextHandle()

        # 3.1
        sendPduHandle = snmpEngine.message_dispatcher.send_pdu(
            snmpEngine,
            transportDomain,
            transportAddress,
            messageProcessingModel,
            securityModel,
            securityName,
            securityLevel,
            contextEngineId,
            contextName,
            pduVersion,
            PDU,
            True,
            timeoutInTicks,
            self.process_response_pdu,
            (sendRequestHandle, cbFun, cbCtx),
        )

        snmpEngine.transport_dispatcher.job_started(id(self))

        self.__pendingReqs[sendPduHandle] = (
            transportDomain,
            transportAddress,
            messageProcessingModel,
            securityModel,
            securityName,
            securityLevel,
            contextEngineId,
            contextName,
            pduVersion,
            origPDU,
            timeoutInTicks,
            retryCount,
            0,
            0,
        )

        debug.logger & debug.FLAG_APP and debug.logger(
            "sendPdu: sendPduHandle %s, timeout %d*10 ms/%d ticks, retry 0 of %d"
            % (sendPduHandle, timeout, timeoutInTicks, retryCount)
        )

        return sendRequestHandle


class GetCommandGenerator(CommandGenerator):
    """SNMP GET command generator."""

    def process_response_varbinds(
        self, snmpEngine: SnmpEngine, sendRequestHandle, errorIndication, PDU, cbCtx
    ):
        """Process SNMP GET response."""
        cbFun, cbCtx = cbCtx

        cbFun(
            snmpEngine,
            sendRequestHandle,
            errorIndication,
            PDU and v2c.apiPDU.get_error_status(PDU) or 0,
            PDU and v2c.apiPDU.get_error_index(PDU, muteErrors=True) or 0,
            PDU and v2c.apiPDU.get_varbinds(PDU) or (),
            cbCtx,
        )

    def send_varbinds(
        self,
        snmpEngine: SnmpEngine,
        targetName,
        contextEngineId,
        contextName,
        varBinds,
        cbFun,
        cbCtx=None,
    ):
        """Send SNMP GET request."""
        reqPDU = v2c.GetRequestPDU()
        v2c.apiPDU.set_defaults(reqPDU)

        v2c.apiPDU.set_varbinds(reqPDU, varBinds)

        return self.send_pdu(
            snmpEngine,
            targetName,
            contextEngineId,
            contextName,
            reqPDU,
            self.process_response_varbinds,
            (cbFun, cbCtx),
        )


class SetCommandGenerator(CommandGenerator):
    """SNMP SET command generator."""

    def process_response_varbinds(
        self, snmpEngine: SnmpEngine, sendRequestHandle, errorIndication, PDU, cbCtx
    ):
        """Process SNMP SET response."""
        cbFun, cbCtx = cbCtx

        cbFun(
            snmpEngine,
            sendRequestHandle,
            errorIndication,
            PDU and v2c.apiPDU.get_error_status(PDU) or 0,
            PDU and v2c.apiPDU.get_error_index(PDU, muteErrors=True) or 0,
            PDU and v2c.apiPDU.get_varbinds(PDU) or (),
            cbCtx,
        )

    def send_varbinds(
        self,
        snmpEngine: SnmpEngine,
        targetName,
        contextEngineId,
        contextName,
        varBinds,
        cbFun,
        cbCtx=None,
    ):
        """Send SNMP SET request."""
        reqPDU = v2c.SetRequestPDU()
        v2c.apiPDU.set_defaults(reqPDU)

        v2c.apiPDU.set_varbinds(reqPDU, varBinds)

        return self.send_pdu(
            snmpEngine,
            targetName,
            contextEngineId,
            contextName,
            reqPDU,
            self.process_response_varbinds,
            (cbFun, cbCtx),
        )


class NextCommandGeneratorSingleRun(CommandGenerator):
    """Single-run SNMP GETNEXT command generator."""

    def process_response_varbinds(
        self, snmpEngine: SnmpEngine, sendRequestHandle, errorIndication, PDU, cbCtx
    ):
        """Process SNMP GETNEXT response."""
        targetName, contextEngineId, contextName, reqPDU, cbFun, cbCtx = cbCtx

        cbFun(
            snmpEngine,
            sendRequestHandle,
            errorIndication,
            PDU and v2c.apiPDU.get_error_status(PDU) or 0,
            PDU and v2c.apiPDU.get_error_index(PDU, muteErrors=True) or 0,
            PDU and v2c.apiPDU.get_varbinds(PDU) or (),
            cbCtx,
        )

    def send_varbinds(
        self,
        snmpEngine: SnmpEngine,
        targetName,
        contextEngineId,
        contextName,
        varBinds,
        cbFun,
        cbCtx=None,
    ):
        """Send SNMP GETNEXT request."""
        reqPDU = v2c.GetNextRequestPDU()
        v2c.apiPDU.set_defaults(reqPDU)

        v2c.apiPDU.set_varbinds(reqPDU, varBinds)

        return self.send_pdu(
            snmpEngine,
            targetName,
            contextEngineId,
            contextName,
            reqPDU,
            self.process_response_varbinds,
            (targetName, contextEngineId, contextName, reqPDU, cbFun, cbCtx),
        )


class NextCommandGenerator(NextCommandGeneratorSingleRun):
    """SNMP GETNEXT command generator."""

    def process_response_varbinds(
        self, snmpEngine: SnmpEngine, sendRequestHandle, errorIndication, PDU, cbCtx
    ):
        """Process SNMP GETNEXT response."""
        targetName, contextEngineId, contextName, reqPDU, cbFun, cbCtx = cbCtx

        if errorIndication:
            cbFun(snmpEngine, sendRequestHandle, errorIndication, 0, 0, (), cbCtx)
            return

        varBinds = v2c.apiPDU.get_varbinds(PDU)

        if v2c.apiPDU.get_error_status(PDU):
            errorIndication, varBinds = None, ()
        elif not varBinds:
            errorIndication, varBinds = errind.emptyResponse, ()

        if not cbFun(
            snmpEngine,
            sendRequestHandle,
            errorIndication,
            v2c.apiPDU.get_error_status(PDU),
            v2c.apiPDU.get_error_index(PDU, muteErrors=True),
            varBinds,
            cbCtx,
        ):
            debug.logger & debug.FLAG_APP and debug.logger(
                "processResponseVarBinds: sendRequestHandle %s, app says to stop walking"
                % sendRequestHandle
            )
            return  # app says enough

        if not varBinds:
            return  # no more objects available


class BulkCommandGeneratorSingleRun(CommandGenerator):
    """Single-run SNMP GETBULK command generator."""

    def process_response_varbinds(
        self, snmpEngine: SnmpEngine, sendRequestHandle, errorIndication, PDU, cbCtx
    ):
        """Process SNMP GETBULK response."""
        (
            targetName,
            nonRepeaters,
            maxRepetitions,
            contextEngineId,
            contextName,
            reqPDU,
            cbFun,
            cbCtx,
        ) = cbCtx

        cbFun(
            snmpEngine,
            sendRequestHandle,
            errorIndication,
            PDU and v2c.apiPDU.get_error_status(PDU) or 0,
            PDU and v2c.apiPDU.get_error_index(PDU, muteErrors=True) or 0,
            PDU and v2c.apiPDU.get_varbinds(PDU) or (),
            cbCtx,
        )

    def send_varbinds(
        self,
        snmpEngine: SnmpEngine,
        targetName,
        contextEngineId,
        contextName,
        nonRepeaters,
        maxRepetitions,
        varBinds,
        cbFun,
        cbCtx=None,
    ):
        """Send SNMP GETBULK request."""
        reqPDU = v2c.GetBulkRequestPDU()
        v2c.apiBulkPDU.set_defaults(reqPDU)

        v2c.apiBulkPDU.set_non_repeaters(reqPDU, nonRepeaters)
        v2c.apiBulkPDU.set_max_repetitions(reqPDU, maxRepetitions)

        v2c.apiBulkPDU.set_varbinds(reqPDU, varBinds)

        return self.send_pdu(
            snmpEngine,
            targetName,
            contextEngineId,
            contextName,
            reqPDU,
            self.process_response_varbinds,
            (
                targetName,
                nonRepeaters,
                maxRepetitions,
                contextEngineId,
                contextName,
                reqPDU,
                cbFun,
                cbCtx,
            ),
        )


class BulkCommandGenerator(BulkCommandGeneratorSingleRun):
    """Bulk SNMP GET command generator."""

    def process_response_varbinds(
        self, snmpEngine: SnmpEngine, sendRequestHandle, errorIndication, PDU, cbCtx
    ):
        """Process SNMP GETBULK response."""
        (
            targetName,
            nonRepeaters,
            maxRepetitions,
            contextEngineId,
            contextName,
            reqPDU,
            cbFun,
            cbCtx,
        ) = cbCtx

        if errorIndication:
            cbFun(snmpEngine, sendRequestHandle, errorIndication, 0, 0, (), cbCtx)
            return

        varBinds = v2c.apiBulkPDU.get_varbinds(PDU)

        if v2c.apiBulkPDU.get_error_status(PDU):
            errorIndication, varBinds = None, ()
        elif not varBinds:
            errorIndication, varBinds = errind.emptyResponse, ()

        if not cbFun(
            snmpEngine,
            sendRequestHandle,
            errorIndication,
            v2c.apiBulkPDU.get_error_status(PDU),
            v2c.apiBulkPDU.get_error_index(PDU, muteErrors=True),
            varBinds,
            cbCtx,
        ):
            debug.logger & debug.FLAG_APP and debug.logger(
                "processResponseVarBinds: sendRequestHandle %s, app says to stop walking"
                % sendRequestHandle
            )
            return  # app says enough

        if not varBinds:
            return  # no more objects available



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# PySNMP MIB module SNMPv2-TM (https://www.pysnmp.com/pysnmp)
# ASN.1 source file://asn1/SNMPv2-TM
# Produced by pysmi-1.5.8 at Sat Nov  2 15:25:46 2024
# On host MacBook-Pro.local platform Darwin version 24.1.0 by user lextm
# Using Python version 3.12.0 (main, Nov 14 2023, 23:52:11) [Clang 15.0.0 (clang-1500.0.40.1)]
#
# IMPORTANT: this file contains customizations

try:
    from socket import inet_ntop, inet_pton, AF_INET
except ImportError:
    from socket import inet_ntoa, inet_aton, AF_INET

    inet_ntop = lambda x, y: inet_ntoa(y)
    inet_pton = lambda x, y: inet_aton(y)


# Import base ASN.1 objects even if this MIB does not use it

(Integer,
 OctetString,
 ObjectIdentifier) = mibBuilder.import_symbols(
    "ASN1",
    "Integer",
    "OctetString",
    "ObjectIdentifier")

(NamedValues,) = mibBuilder.import_symbols(
    "ASN1-ENUMERATION",
    "NamedValues")
(ConstraintsIntersection,
 ConstraintsUnion,
 SingleValueConstraint,
 ValueRangeConstraint,
 ValueSizeConstraint) = mibBuilder.import_symbols(
    "ASN1-REFINEMENT",
    "ConstraintsIntersection",
    "ConstraintsUnion",
    "SingleValueConstraint",
    "ValueRangeConstraint",
    "ValueSizeConstraint")

# Import SMI symbols from the MIBs this MIB depends on

(ModuleCompliance,
 NotificationGroup) = mibBuilder.import_symbols(
    "SNMPv2-CONF",
    "ModuleCompliance",
    "NotificationGroup")

(Bits,
 Counter32,
 Counter64,
 Gauge32,
 Integer32,
 IpAddress,
 ModuleIdentity,
 MibIdentifier,
 NotificationType,
 ObjectIdentity,
 MibScalar,
 MibTable,
 MibTableRow,
 MibTableColumn,
 TimeTicks,
 Unsigned32,
 iso,
 snmpDomains,
 snmpModules,
 snmpProxys) = mibBuilder.import_symbols(
    "SNMPv2-SMI",
    "Bits",
    "Counter32",
    "Counter64",
    "Gauge32",
    "Integer32",
    "IpAddress",
    "ModuleIdentity",
    "MibIdentifier",
    "NotificationType",
    "ObjectIdentity",
    "MibScalar",
    "MibTable",
    "MibTableRow",
    "MibTableColumn",
    "TimeTicks",
    "Unsigned32",
    "iso",
    "snmpDomains",
    "snmpModules",
    "snmpProxys")

(DisplayString,
 TextualConvention) = mibBuilder.import_symbols(
    "SNMPv2-TC",
    "DisplayString",
    "TextualConvention")


# MODULE-IDENTITY

snmpv2tm = ModuleIdentity(
    (1, 3, 6, 1, 6, 3, 19)
)
if mibBuilder.loadTexts:
    snmpv2tm.setRevisions(
        ("2002-10-16 00:00",
         "1996-01-01 00:00",
         "1993-04-01 00:00")
    )
if mibBuilder.loadTexts:
    snmpv2tm.setLastUpdated("200210160000Z")
if mibBuilder.loadTexts:
    snmpv2tm.setOrganization("IETF SNMPv3 Working Group")
if mibBuilder.loadTexts:
    snmpv2tm.setContactInfo("WG-EMail: snmpv3@lists.tislabs.com Subscribe: snmpv3-request@lists.tislabs.com Co-Chair: Russ Mundy Network Associates Laboratories postal: 15204 Omega Drive, Suite 300 Rockville, MD 20850-4601 USA EMail: mundy@tislabs.com phone: +1 301 947-7107 Co-Chair: David Harrington Enterasys Networks postal: 35 Industrial Way P. O. Box 5005 Rochester, NH 03866-5005 USA EMail: dbh@enterasys.com phone: +1 603 337-2614 Editor: Randy Presuhn BMC Software, Inc. postal: 2141 North First Street San Jose, CA 95131 USA EMail: randy_presuhn@bmc.com phone: +1 408 546-1006")
if mibBuilder.loadTexts:
    snmpv2tm.setDescription("The MIB module for SNMP transport mappings. Copyright (C) The Internet Society (2002). This version of this MIB module is part of RFC 3417; see the RFC itself for full legal notices. ")


# Types definitions


# TEXTUAL-CONVENTIONS



class SnmpUDPAddress(TextualConvention, OctetString):
    status = "current"
    displayHint = "1d.1d.1d.1d/2d"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(6, 6),
    )
    fixed_length = 6

    if mibBuilder.loadTexts:
        description = "Represents a UDP over IPv4 address: octets contents encoding 1-4 IP-address network-byte order 5-6 UDP-port network-byte order "

    def prettyIn(self, value):
        if isinstance(value, tuple):
            # Wild hack -- need to implement TextualConvention.prettyIn
            value = (
                inet_pton(AF_INET, value[0])
                + bytes((((value[1] >> 8) & 0xFF),))
                + bytes(((value[1] & 0xFF),))
            )
        return OctetString.prettyIn(self, value)

    # Socket address syntax coercion
    def __asSocketAddress(self):
        if not hasattr(self, "__tuple_value"):
            v = self.asOctets()
            self.__tuple_value = (
                inet_ntop(AF_INET, v[:4]),
                v[4] << 8 | v[5],
            )
        return self.__tuple_value

    def __iter__(self):
        return iter(self.__asSocketAddress())

    def __getitem__(self, item):
        return self.__asSocketAddress()[item]



class SnmpOSIAddress(TextualConvention, OctetString):
    status = "current"
    displayHint = "*1x:/1x:"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(1, 1),
        ValueSizeConstraint(4, 85),
    )

    if mibBuilder.loadTexts:
        description = "Represents an OSI transport-address: octets contents encoding 1 length of NSAP 'n' as an unsigned-integer (either 0 or from 3 to 20) 2..(n+1) NSAP concrete binary representation (n+2)..m TSEL string of (up to 64) octets "


class SnmpNBPAddress(TextualConvention, OctetString):
    status = "current"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(3, 99),
    )

    if mibBuilder.loadTexts:
        description = "Represents an NBP name: octets contents encoding 1 length of object 'n' as an unsigned integer 2..(n+1) object string of (up to 32) octets n+2 length of type 'p' as an unsigned integer (n+3)..(n+2+p) type string of (up to 32) octets n+3+p length of zone 'q' as an unsigned integer (n+4+p)..(n+3+p+q) zone string of (up to 32) octets For comparison purposes, strings are case-insensitive. All strings may contain any octet other than 255 (hex ff)."


class SnmpIPXAddress(TextualConvention, OctetString):
    status = "current"
    displayHint = "4x.1x:1x:1x:1x:1x:1x.2d"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(12, 12),
    )
    fixed_length = 12

    if mibBuilder.loadTexts:
        description = "Represents an IPX address: octets contents encoding 1-4 network-number network-byte order 5-10 physical-address network-byte order 11-12 socket-number network-byte order "


# MIB Managed Objects in the order of their OIDs

_SnmpUDPDomain_ObjectIdentity = ObjectIdentity
snmpUDPDomain = _SnmpUDPDomain_ObjectIdentity(
    (1, 3, 6, 1, 6, 1, 1)
)
if mibBuilder.loadTexts:
    snmpUDPDomain.setStatus("current")
if mibBuilder.loadTexts:
    snmpUDPDomain.setDescription("The SNMP over UDP over IPv4 transport domain. The corresponding transport address is of type SnmpUDPAddress.")
_SnmpCLNSDomain_ObjectIdentity = ObjectIdentity
snmpCLNSDomain = _SnmpCLNSDomain_ObjectIdentity(
    (1, 3, 6, 1, 6, 1, 2)
)
if mibBuilder.loadTexts:
    snmpCLNSDomain.setStatus("current")
if mibBuilder.loadTexts:
    snmpCLNSDomain.setDescription("The SNMP over CLNS transport domain. The corresponding transport address is of type SnmpOSIAddress.")
_SnmpCONSDomain_ObjectIdentity = ObjectIdentity
snmpCONSDomain = _SnmpCONSDomain_ObjectIdentity(
    (1, 3, 6, 1, 6, 1, 3)
)
if mibBuilder.loadTexts:
    snmpCONSDomain.setStatus("current")
if mibBuilder.loadTexts:
    snmpCONSDomain.setDescription("The SNMP over CONS transport domain. The corresponding transport address is of type SnmpOSIAddress.")
_SnmpDDPDomain_ObjectIdentity = ObjectIdentity
snmpDDPDomain = _SnmpDDPDomain_ObjectIdentity(
    (1, 3, 6, 1, 6, 1, 4)
)
if mibBuilder.loadTexts:
    snmpDDPDomain.setStatus("current")
if mibBuilder.loadTexts:
    snmpDDPDomain.setDescription("The SNMP over DDP transport domain. The corresponding transport address is of type SnmpNBPAddress.")
_SnmpIPXDomain_ObjectIdentity = ObjectIdentity
snmpIPXDomain = _SnmpIPXDomain_ObjectIdentity(
    (1, 3, 6, 1, 6, 1, 5)
)
if mibBuilder.loadTexts:
    snmpIPXDomain.setStatus("current")
if mibBuilder.loadTexts:
    snmpIPXDomain.setDescription("The SNMP over IPX transport domain. The corresponding transport address is of type SnmpIPXAddress.")
_Rfc1157Proxy_ObjectIdentity = ObjectIdentity
rfc1157Proxy = _Rfc1157Proxy_ObjectIdentity(
    (1, 3, 6, 1, 6, 2, 1)
)
_Rfc1157Domain_ObjectIdentity = ObjectIdentity
rfc1157Domain = _Rfc1157Domain_ObjectIdentity(
    (1, 3, 6, 1, 6, 2, 1, 1)
)
if mibBuilder.loadTexts:
    rfc1157Domain.setStatus("deprecated")
if mibBuilder.loadTexts:
    rfc1157Domain.setDescription("The transport domain for SNMPv1 over UDP over IPv4. The corresponding transport address is of type SnmpUDPAddress.")

# Managed Objects groups


# Notification objects


# Notifications groups


# Agent capabilities


# Module compliance


# Export all MIB objects to the MIB builder

mibBuilder.export_symbols(
    "SNMPv2-TM",
    **{"SnmpUDPAddress": SnmpUDPAddress,
       "SnmpOSIAddress": SnmpOSIAddress,
       "SnmpNBPAddress": SnmpNBPAddress,
       "SnmpIPXAddress": SnmpIPXAddress,
       "snmpUDPDomain": snmpUDPDomain,
       "snmpCLNSDomain": snmpCLNSDomain,
       "snmpCONSDomain": snmpCONSDomain,
       "snmpDDPDomain": snmpDDPDomain,
       "snmpIPXDomain": snmpIPXDomain,
       "rfc1157Proxy": rfc1157Proxy,
       "rfc1157Domain": rfc1157Domain,
       "snmpv2tm": snmpv2tm,
       "PYSNMP_MODULE_ID": snmpv2tm}
)



import pytest
from pysnmp.hlapi.v3arch.asyncio import *
from pysnmp.proto.rfc1902 import OctetString
from pysnmp.smi.rfc1902 import ObjectIdentity, ObjectType
from tests.agent_context import AGENT_PORT, AgentContextManager


@pytest.mark.asyncio
async def test_v2_set():
    async with AgentContextManager():
        with SnmpEngine() as snmpEngine:
            errorIndication, errorStatus, errorIndex, varBinds = await set_cmd(
                snmpEngine,
                CommunityData("public"),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ContextData(),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysLocation", 0), "Shanghai"),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysLocation.0"
            assert varBinds[0][1].prettyPrint() == "Shanghai"
            assert isinstance(varBinds[0][1], OctetString)
